"""
Shared utilities accessible to all components
"""

# Standard
from typing import Dict, Optional, Union
import base64
import copy
import re

# First Party
import alog

# Local
from . import constants

# First party
from oper8 import Component, Session, assert_config

log = alog.use_channel("UTIL")


def b64_secret(val):
    if isinstance(val, str):
        val = val.encode("utf-8")
    return base64.b64encode(val).decode("utf-8")


def b64_secret_decode(val):
    if isinstance(val, str):
        val = val.encode("utf-8")
    return base64.b64decode(val).decode("utf-8")


def get_resource_cluster_name(
    resource_name: str,
    component: Component,
    session: Session,
):
    """Common helper function to get the name a given kubernetes resource should
    use when deployed to the cluster.

    Args:
        resource_name:  str
            The raw name for the resource (e.g. sireg-secret)
        component:  Union[Component, str]
            The component (or component name) that owns this resource
        session:  Session
            The session for the current reconciliation deploy

    Returns:
        resource_cluster_name:  str
            The resource name with appropriate scoping and truncation added
    """
    if is_global(component, session):
        log.debug2(
            "Applying global name logic to [%s] for component [%s]",
            resource_name,
            component,
        )
        return session.get_truncated_name(resource_name)
    return session.get_scoped_name(resource_name)


def get_replicas(
    session: Session,
    component_name: str,
    unscoped_name: str,
    force: bool = False,
    replicas_override: Union[int, None] = None,
) -> Union[int, None]:
    """
    Get the replica count for the given resource.

    This function consolidates logic for getting replicas for all components in
    the application. It allows replicas to be conditionally set only when needed
    to avoid thrashing with HPAs.

    Args:
        session: Session
            The current deploy session
        component_name: str
            The name of the component to get replicas for
        unscoped_name: str
            The external name of the deployment without scoping
        force: bool
            If True, the state of the cluster will not be checked
        replicas_override: int or None
            An override value to use in place of the normal config-based value

    Returns:
        replicas: int or None
            If replicas should not be set for this resource, None is returned,
            otherwise the number of replicas is returned based on the t-shirt
            size for the instance.
    """

    # Fetch the current state of the deployment
    if not force:
        name = get_resource_cluster_name(
            resource_name=unscoped_name,
            component=component_name,
            session=session,
        )
        success, content = session.get_object_current_state(
            kind="Deployment",
            name=name,
            api_version="apps/v1",
        )
        assert success, f"Failed to look up state for [{name}]"

        # Check the current content to see if this is a t-shirt size change
        if content is not None:
            # Fetch the current replica count. We'll reuse this if there's no
            # reason to change
            replicas = content.get("spec", {}).get("replicas")

            # If we found replicas, check for t-shirt size change
            if replicas is None:
                log.debug("No replicas found for [%s]. Using config.".name)
            else:
                assert isinstance(replicas, int), "Replicas is not an int!"
                current_size = session.spec.size
                deployed_size = (
                    content.get("metadata", {}).get("labels", {}).get("instance-size")
                )
                if replicas == 0 and not session.spec.get("backup", {}).get(
                    "offlineQuiesce", False
                ):
                    log.debug(
                        "Found [%s] with size [%s] and offlineQuiesce off. Need "
                        "to scale up from [%s] replicas.",
                        name,
                        current_size,
                        replicas,
                    )
                elif current_size == deployed_size:
                    log.debug(
                        "Found [%s] with size [%s]. Not changing replicas from [%s].",
                        name,
                        current_size,
                        replicas,
                    )
                    return replicas
                else:
                    log.debug(
                        "Found t-shirt size change for [%s] from [%s -> %s]",
                        name,
                        deployed_size,
                        current_size,
                    )

    # Look up the replicas based on the t-shirt size
    size = session.spec.size
    replica_map = session.config.get("replicas", {}).get(size, {})
    replicas = replicas_override or replica_map.get(component_name)
    log.debug3("Replica map for [%s]: %s", size, replica_map)
    assert_config(
        replicas is not None,
        f"No replicas for [{component_name}] available for size [{size}]",
    )
    return replicas


def is_global(component: Union[Component, str], session: Session) -> bool:
    """Determine if the given component is global in this deployment

    Args:
        component:  Union[Component, str]
            The component to fetch the slot name for
        session:  Session
            The session for the current deployment

    Returns:
        is_global:  bool
            True if the given component is global!
    """
    component_name = component if isinstance(component, str) else component.name
    return session.config.get(component_name, {}).get(constants.GLOBAL_SLOT, False)


def get_deploy_labels(session, base_labels=None):
    """Get labels for a Deployment resource on top of the standard base labels"""
    # Shallow copy is fine here since labels are one-level deep and only strings
    deploy_labels = copy.copy(base_labels or {})
    deploy_labels["instance-size"] = session.spec.size
    return deploy_labels


def camelcase_to_snake_case(x):
    if isinstance(x, str):
        return re.sub(r"(?<!^)(?=[A-Z])", "_", x).lower()
    elif isinstance(x, list):
        return [y if isinstance(y, str) else camelcase_to_snake_case(y) for y in x]
    elif isinstance(x, dict):
        return {
            camelcase_to_snake_case(k): v
            if isinstance(v, str)
            else camelcase_to_snake_case(v)
            for k, v in x.items()
        }
    else:
        return x


def snake_case_to_camelcase(x):
    if isinstance(x, str):
        return (
            x[0].lower() + "".join([world.capitalize() for world in x.split("_")])[1:]
        )
    elif isinstance(x, list):
        return [y if isinstance(y, str) else snake_case_to_camelcase(y) for y in x]
    elif isinstance(x, dict):
        return {
            snake_case_to_camelcase(k): v
            if isinstance(v, str)
            else snake_case_to_camelcase(v)
            for k, v in x.items()
        }
    else:
        return x


def get_slot_name(component: Union[Component, str], session: Session) -> str:
    """Get the slot name for the given component in the current deployment

    Args:
        component:  Union[Component, str]
            The component to fetch the slot name for
        session:  DeploySession
            The session for the current deployment

    Returns:
        slot_name:  str
            The string name of the slot where the given component will live for
            this deployment. For global components, the static global slot name
            is returned
    """
    if not is_global(component, session):
        return session.name
    return constants.GLOBAL_SLOT


def get_labels(
    cluster_name: str,
    session: Session,
    component_name: Optional[str] = None,
) -> Dict[str, str]:
    """Common utility for fetching the set of metadata.labels for a given resource.
    Args:
        cluster_name:  str
            The name of the resource as it will be applied to the cluster
            including any scoping applied by get_resource_cluster_name
        session:  DeploySession
            The session for the current deployment
        component_name:  str
            The name of the component that manages this resource.
            NOTE: This argument is optional for backwards compatibility,
                but should always be provided to ensure accurate labels!
    Returns:
        labels:  Dict[str, str]
            The full set of labels to use for the given resource
    """
    labels = {
        "app": cluster_name,
        "app.kubernetes.io/managed-by": "Oper8",
        "app.kubernetes.io/instance": session.name,
    }
    if component_name:
        labels["component"] = component_name
    if slot_name := get_slot_name(component_name, session):
        labels["slot"] = slot_name

    # Add user-specified labels from the CR's spec.labels field
    user_labels = session.spec.labels or {}
    labels.update(user_labels)

    return labels


def metadata_defaults(
    cluster_name: str,
    session: Session,
    **kwargs,
) -> dict:
    """This function will create the metadata object given the external name for
    a resource. The external name should be created using
    common.get_resource_external_name. These functions are separate because the
    external name is often needed independently, so it will be pre-computed at
    the start of most components.

    Args:
        cluster_name:  str
            The fully scoped and truncated name that the resource will use in
            the cluster (metadata.name)
        session:  DeploySession
            The session for the current reconciliation deploy

    Returns:
        metadata:  dict
            The constructed metadata dict
    """
    # NOTE: For the time being, there are no defaults injected here, but we will
    #   retain the abstraction function so that we can add defaulting
    #   functionality (e.g. multi-namespace deployments) without touching every
    #   file.
    return {"name": cluster_name, **kwargs}


def from_string_or_number(value: Union[int, str, float]) -> Union[int, str, float]:
    """Handle strings or numbers for fields that can be either string or numeric

    Args:
        value: Union[str, int, float]
            Quantity type that can be in numeric or string form (e.g. resources)

    Returns:
        formatted_value: Union[str, int, float]
            The value formatted as the correct type
    """
    # By default no conversion is needed
    formatted_value = value

    # If it's a string, try converting it to an int, then a float
    if isinstance(value, str):
        for target_type in [int, float]:
            try:
                formatted_value = target_type(value)
                break
            except ValueError:
                pass

    return formatted_value


def mount_mode(octal_val):
    """This helper gets the decimal version of an octal representation of file
    permissions used for a volume mount.

    Args:
        octal_val:  int or str
            The number as octal (e.g. 755 or "0755")

    Returns:
        decimal_val:  int
            The decimal integer value corresponding to the given octal value
            which can be used in VolumeMount's default_mode field
    """
    return int(str(octal_val), 8)



"""
Sharted utils for testing postgres
"""

# First Party
import aconfig

# Local
from oper8.test_helpers.helpers import TEST_INSTANCE_NAME
from oper8.test_helpers.oper8x_helpers import set_secret_data
from oper8.x.datastores.postgres.factory import PostgresFactory
from oper8.x.utils import common, constants

## Globals #####################################################################

SECRET_NAME_UNSCOPED_AUTH = "postgres-admin-auth"
SECRET_NAME_UNSCOPED_TLS = "postgres-ca"
# Used for testing/dry run purposes
SECRET_NAME_UNSCOPED_URI = "postgres-uri"


class AuthSecretKeys:
    USING_SECRET = "USING_SECRET"
    USERNAME = "username"
    PASSWORD = "password"
    PGPASS = "pgpass"
    PG_REPLICATION_USER = "PG_REPLICATION_USER"
    PG_REPLICATION_PASSWORD = "PG_REPLICATION_PASSWORD"
    ALL_KEYS = [
        USERNAME,
        PASSWORD,
        PGPASS,
    ]


class TlsSecretKeys:
    CERT = "ca.crt"
    KEY = "ca.key"
    CA_CERT = "ca.crt"
    ALL_KEYS = [
        CERT,
        KEY,
        CA_CERT,
    ]


class UriSecretKeys:
    HOSTNAME = "hostname"
    PORT = "port"


TEST_POSTGRES_AUTH = {
    AuthSecretKeys.USING_SECRET: common.b64_secret("fake-secret"),
    AuthSecretKeys.USERNAME: common.b64_secret("fake-user"),
    AuthSecretKeys.PASSWORD: common.b64_secret("fake-password"),
    AuthSecretKeys.PGPASS: common.b64_secret("fake-pgpass"),
    AuthSecretKeys.PG_REPLICATION_USER: common.b64_secret("fake-replication-user"),
    AuthSecretKeys.PG_REPLICATION_PASSWORD: common.b64_secret(
        "fake-replication-password"
    ),
}

TEST_POSTGRES_TLS = {
    TlsSecretKeys.CERT: common.b64_secret("fake-tls-cert"),
    TlsSecretKeys.CA_CERT: common.b64_secret("fake-tls-ca-cert"),
    TlsSecretKeys.KEY: common.b64_secret("fake-tls-key"),
}


TEST_POSTGRES_URI = {
    UriSecretKeys.HOSTNAME: common.b64_secret("fake-uri-hostname"),
    UriSecretKeys.PORT: common.b64_secret("51423"),
}

POSTGRES_PROVIDED_CONNECTION_DEPLOY_CONFIG_OVERRIDES = {
    constants.SPEC_DATASTORES: {
        PostgresFactory.DATASTORE_TYPE: {
            constants.SPEC_DATASTORE_CONNECTION: {
                "uri_secret": f"{TEST_INSTANCE_NAME}-{SECRET_NAME_UNSCOPED_URI}",
                "uri_secret_hostname_field": UriSecretKeys.HOSTNAME,
                "uri_secret_port_field": UriSecretKeys.PORT,
                "auth_secret_name": f"{TEST_INSTANCE_NAME}-{SECRET_NAME_UNSCOPED_AUTH}",
                "auth_secret_username_field": AuthSecretKeys.USERNAME,
                "auth_secret_password_field": AuthSecretKeys.PASSWORD,
                "tls_secret_name": f"{TEST_INSTANCE_NAME}-{SECRET_NAME_UNSCOPED_TLS}",
                "tls_secret_cert_field": TlsSecretKeys.CERT,
            }
        }
    }
}


def set_postgres_secrets(session):
    set_postgres_auth_secret(session)
    set_postgres_tls_secret(session)
    set_postgres_uri_secret(session)


def set_postgres_auth_secret(session, override=None):
    secret_data = override or TEST_POSTGRES_AUTH
    set_secret_data(
        session,
        SECRET_NAME_UNSCOPED_AUTH,
        secret_data,
    )


def set_postgres_tls_secret(session, override=None):
    secret_data = override or TEST_POSTGRES_TLS
    set_secret_data(
        session,
        SECRET_NAME_UNSCOPED_TLS,
        data=secret_data,
    )


def set_postgres_uri_secret(session, override=None):
    secret_data = override or TEST_POSTGRES_URI
    set_secret_data(session, SECRET_NAME_UNSCOPED_URI, secret_data)


def get_spec_overrides():
    return {
        constants.SPEC_DATASTORES: {"postgres": {"storageClassName": "test-storage"}}
    }



"""
This test simulates the sequence of events that will be needed when an
implementation of ITlsContext needs to wait on a downstream operator (such as
certificate manager) to create the TLS content.
"""

# Standard
import abc

# First Party
import alog

# Local
from oper8.test_helpers.helpers import MockDeployManager, MockTopApp, setup_session
from oper8.x.oper8x_component import Oper8xComponent
from oper8.x.utils import common, tls
from oper8.x.utils.tls_context.factory import register_tls_context_type
from oper8.x.utils.tls_context.interface import ITlsContext
from tests.x.utils.tls_context.util import make_server_component_class

## Helpers #####################################################################

log = alog.use_channel("TEST")


class FakeCMTlsContext(ITlsContext):

    TYPE_LABEL = "fakecm"

    def __init__(self, session, config):
        super().__init__(session, config)
        self.iteration = 0

        # Add the CA component
        log.debug("Setting up CA component")
        self._ca_component = DelayedCaTlsComponent(
            session, config.get("ca_wait_iterations", 1)
        )

    def get_client_cert(
        self,
        client_component,
        encode=True,
    ):
        # Declare a dependency for the client component on the CA component
        self.session.add_component_dependency(client_component, self._ca_component)

        # Look for the CA component's content to see if it exists
        _, content = self.session.get_object_current_state(
            kind="Secret",
            name=self._ca_component.SECRET_NAME,
        )

        # If it exists, fetch the CA crt and optionally decode
        if content is not None:
            log.debug("Found valid CA content")
            ca_crt = content["data"][self._ca_component.CA_CRT_FIELD]
            if not encode:
                ca_crt = common.b64_secret_decode(ca_crt)
            return ca_crt

        # If it doesn't exist, return None
        log.debug("Still waiting on valid CA content")
        return None

    def get_server_key_cert_pair(
        self,
        server_component,
        san_hostnames_list=None,
        san_ip_list=None,
        encode=True,
        intermediate_ca=False,
    ):
        # Add the component that will provision the server key/cert pair (if already not created)
        # in tests we share deploy manager, i.e. we also share deployment id, contexts and components
        #  between steps. Thus registering the component just once per server component
        if "delayed-tls-{server_component.name}" not in self.session.get_components():
            log.debug("Setting up server tls component for %s", server_component.name)
            tls_component = make_delayed_server_component(
                server_component.name,
                self.session,
                self.config.get("server_wait_iterations", 1),
            )
            self.session.add_component_dependency(server_component, tls_component)
            self.session.add_component_dependency(tls_component, self._ca_component)

        # Check the cluster to see if it's been provisioned yet
        _, content = self.session.get_object_current_state(
            kind="Secret",
            name=tls_component.SECRET_NAME,
        )

        # If it exists, fetch the key/cert pair and optionally decode
        if content is not None:
            log.debug("Found server tls content for %s", server_component.name)
            key = content["data"][tls_component.KEY_FIELD]
            crt = content["data"][tls_component.CRT_FIELD]
            if not encode:
                key = common.b64_secret_decode(key)
                crt = common.b64_secret_decode(crt)
            return key, crt

        # If it doesn't exist, return None
        log.debug("Still waiting on server tls content for %s", server_component.name)
        return None, None


# Register the dummy implementation
register_tls_context_type(FakeCMTlsContext)


class DelayedTlsComponent(Oper8xComponent):
    ITERATION_SECRET_BASE_NAME = "iterations-secret"

    def __init__(self, session, wait_iterations):
        super().__init__(session=session)
        self.wait_iterations = wait_iterations

    def build_chart(self, *_, **__):

        # Get the current iteration from the existing secret
        iteration_secret_name = f"{self.ITERATION_SECRET_BASE_NAME}-{self.name}"
        _, content = self.session.get_object_current_state(
            kind="Secret", name=iteration_secret_name
        )
        current_iteration = 0
        if content is not None:
            current_iteration = int(
                common.b64_secret_decode(content["data"]["iteration"])
            )

        # If we hit the iteration threshold, add the generated TLS content
        log.debug2(
            "[%s] current iteration: %d, wait iterations: %d",
            self,
            current_iteration,
            self.wait_iterations,
        )
        if current_iteration == self.wait_iterations:
            self.add_tls_secret()
        else:
            log.debug2("[%s] Waiting to add TLS secret", self)

        # Increment the iteration
        current_iteration += 1
        self.add_resource(
            name=iteration_secret_name,
            obj=dict(
                kind="Secret",
                apiVersion="v1",
                metadata=dict(name=iteration_secret_name),
                data={"iteration": common.b64_secret(str(current_iteration))},
            ),
        )

    @abc.abstractmethod
    def add_tls_secret(self):
        """Will be implemented below, once for CA and once for derived server"""


class DelayedCaTlsComponent(DelayedTlsComponent):
    name = "delayed-tls-ca"

    SECRET_NAME = "ca-secret"
    CA_KEY_FIELD = "key.pem"
    CA_CRT_FIELD = "crt.pem"

    def add_tls_secret(self):
        log.debug("Creating TLS secret for %s", self)

        # Generate key and CA cert
        key, ca_key = tls.generate_key(encode=True)
        ca_crt = tls.generate_ca_cert(key, encode=True)

        # Create the secret
        self.add_resource(
            name=self.SECRET_NAME,
            obj=dict(
                kind="Secret",
                apiVersion="v1",
                metadata=dict(name=self.SECRET_NAME),
                data={
                    self.CA_KEY_FIELD: ca_key,
                    self.CA_CRT_FIELD: ca_crt,
                },
            ),
        )


def make_delayed_server_tls_secret_name(server_name):
    return f"{server_name}-tls"


def make_delayed_server_component(server_name, session, wait_iterations):
    class DelayedServerTlsComponent(DelayedTlsComponent):
        name = f"delayed-tls-{server_name}"

        SECRET_NAME = make_delayed_server_tls_secret_name(server_name)
        KEY_FIELD = "key.pem"
        CRT_FIELD = "crt.pem"

        def add_tls_secret(self):
            log.debug("Creating TLS secret for %s", self)

            # Get the CA from the cluster
            _, ca_content = self.session.get_object_current_state(
                kind="Secret",
                name=DelayedCaTlsComponent.SECRET_NAME,
            )
            assert ca_content is not None
            key = tls.parse_private_key_pem(
                common.b64_secret_decode(
                    ca_content["data"][DelayedCaTlsComponent.CA_KEY_FIELD]
                )
            )

            # Derive the server key/cert pair
            key, crt = tls.generate_derived_key_cert_pair(key, [], encode=True)

            # Create the secret
            self.add_resource(
                name=self.SECRET_NAME,
                obj=dict(
                    kind="Secret",
                    apiVersion="v1",
                    metadata=dict(name=self.SECRET_NAME),
                    data={
                        self.KEY_FIELD: key,
                        self.CRT_FIELD: crt,
                    },
                ),
            )

    return DelayedServerTlsComponent(session, wait_iterations)


def has_secret(session, name):
    return session.deploy_manager.has_obj(
        kind="Secret",
        name=name,
        namespace=session.namespace,
    )


## Tests #######################################################################


def test_server_precondition_assert():
    """Test that a server component which needs to wait for TLS content
    correctly waits, but proceeds once the content is available
    """
    dm = MockDeployManager()
    app_config_overrides = {
        "tls": {
            "type": FakeCMTlsContext.TYPE_LABEL,
            "ca_wait_iterations": 1,
            "server_wait_iterations": 1,
        }
    }

    # Define the server component class
    ServerComponent = make_server_component_class()

    # Do the first pass. We want to ensure that it does not get past the wait on
    # the TLS data
    session = setup_session(
        deploy_manager=dm,
        app_config=app_config_overrides,
    )
    with alog.ContextLog(log.debug, "---- [FIRST PASS] ----"):
        # Set up a top-level application that will manage this component
        app = MockTopApp(session.config, [ServerComponent])

        # Roll it out
        app.do_rollout(session)

        # Make sure that the component's secret was not created and that none of
        # the TLS data secrets were created
        assert not has_secret(session, DelayedCaTlsComponent.SECRET_NAME)
        assert not has_secret(
            session, make_delayed_server_tls_secret_name(ServerComponent.name)
        )
        assert not has_secret(session, ServerComponent.get_secret_name())

    # Do the second pass. The TLS "backend" should provision the data this time,
    # but since that happens at deploy time, the server will still not see it,
    # so it will again wait on the precondition.
    session = setup_session(
        deploy_manager=dm,
        app_config=app_config_overrides,
    )
    with alog.ContextLog(log.debug, "---- [SECOND PASS] ----"):
        # Set up a top-level application that will manage this component
        app = MockTopApp(session.config, [ServerComponent])

        # Roll it out
        app.do_rollout(session)

        # Make sure that the upstream secrets have been provisioned, but not
        # the server's secret yet
        assert has_secret(session, DelayedCaTlsComponent.SECRET_NAME)
        assert has_secret(
            session, make_delayed_server_tls_secret_name(ServerComponent.name)
        )
        assert not has_secret(session, ServerComponent.get_secret_name())

    # Do the third pass. Now, the TLS data exists in the cluster before the
    # rollout starts, so the server can proceed.
    session = setup_session(
        deploy_manager=dm,
        app_config=app_config_overrides,
    )
    with alog.ContextLog(log.debug, "---- [THIRD PASS] ----"):
        # Set up a top-level application that will manage this component
        app = MockTopApp(session.config, [ServerComponent])

        # Roll it out
        app.do_rollout(session)

        # Make sure that the server's secret is there
        assert has_secret(session, DelayedCaTlsComponent.SECRET_NAME)
        assert has_secret(
            session, make_delayed_server_tls_secret_name(ServerComponent.name)
        )
        assert has_secret(session, ServerComponent.get_secret_name())



"""
Test the TLS context utility functionality
"""

# Standard
from contextlib import closing, contextmanager
import multiprocessing
import os
import random
import socket
import tempfile
import time

# Third Party
from flask import Flask
import requests

# First Party
import alog

# Local
from oper8 import Component, component
from oper8.test_helpers.helpers import (
    MockDeployManager,
    configure_logging,
    setup_session,
)
from oper8.test_helpers.oper8x_helpers import set_secret_data, set_tls_ca_secret
from oper8.x.utils import common, tls_context
from oper8.x.utils.tls_context.factory import (
    _TlsContextSingletonFactory,
    get_tls_context,
    register_tls_context_type,
)
from oper8.x.utils.tls_context.internal import InternalCaComponent, InternalTlsContext

## Helpers #####################################################################

configure_logging()
log = alog.use_channel("TEST")


@component(name="Stub")
class StubChart(Component):
    def build_chart(self, session):
        log.debug("Constructing %s", self)
        self.doit(session=session, scope=self)


def reset():
    _TlsContextSingletonFactory._instance = None
    _TlsContextSingletonFactory._instance_deploy_id = None


def make_stub_chart(session, doit, name=None):
    chart_class = StubChart
    if name is not None:
        component_name = name

        class DerivedChart(StubChart):
            name = component_name

        chart_class = DerivedChart

    chart = chart_class(session=session)
    setattr(chart, "doit", doit)
    return chart


def port_open(port):
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
        return sock.connect_ex(("127.0.0.1", port)) != 0


def random_port():
    return int(random.uniform(12345, 55555))


def get_available_port():
    port = random_port()
    while not port_open(port):
        port = random_port
    return port


def get_internal_ctxt_ca_pair(session):
    ctxt = get_tls_context(session)
    return ctxt._component._ca_key_pem, ctxt._component._ca_crt_pem


INTERNAL_TLS_OVERRIDES = {"tls": {"type": "internal"}}


## State Management ############################################################
#
# These tests exercise the global state management to ensure that the core CA
# key and cert are generated and reused correctly
##


def test_generate_when_not_found_get_client():
    """Test that the CA key/cert are generated when not found in the cluster and
    a client cert is requested
    """

    # Create the chart
    def doit(session, scope):
        crt = tls_context.get_client_cert(session, scope, encode=False)
        assert crt is not None
        assert crt == get_internal_ctxt_ca_pair(session)[1]

    # Trigger the render
    reset()
    session = setup_session(
        app_config=INTERNAL_TLS_OVERRIDES,
    )
    chart = make_stub_chart(session, doit)
    chart.to_config(session)


def test_generate_when_not_found_get_server():
    """Test that the CA key/cert are generated when not found in the cluster and
    a server key/cert pair is requested
    """

    # Create the chart
    def doit(session, scope):
        tls_context.request_server_key_cert_pair(session, scope, [])
        key, crt = tls_context.get_server_key_cert_pair(session, scope, encode=False)

        # Make sure that the cached values are populated and the generated key
        # and cert do not match the cached CA
        assert key is not None
        assert crt is not None
        ca_key, ca_crt = get_internal_ctxt_ca_pair(session)
        assert ca_key is not None
        assert ca_crt is not None
        assert key != ca_key
        assert crt != ca_crt

    # Trigger the render
    reset()
    session = setup_session(
        app_config=INTERNAL_TLS_OVERRIDES,
    )
    chart = make_stub_chart(session, doit)
    chart.to_config(session)


def test_generate_when_found_but_missing_keys():
    """Test that the CA key/cert are regenerated when not found in the cluster
    but the secret is missing values
    """
    reset()
    session = setup_session(
        app_config=INTERNAL_TLS_OVERRIDES,
        deploy_manager=MockDeployManager(),
    )
    # Pre-populate the secret but don't add the keys
    set_secret_data(session, InternalCaComponent.CA_SECRET_NAME, data={})
    chart = InternalCaComponent(session)

    # Trigger the render and make sure the generated secret matches the
    # cached CA content
    configs = chart.to_config(session)
    assert len(configs) == 1
    secret_crt = common.b64_secret_decode(
        configs[0]["data"][InternalCaComponent.CA_CRT_FILENAME]
    )
    secret_key = common.b64_secret_decode(
        configs[0]["data"][InternalCaComponent.CA_KEY_FILENAME]
    )
    assert (secret_key, secret_crt) == (chart._ca_key_pem, chart._ca_crt_pem)


def test_reuse_when_found_get_client():
    """Test that the CA key/cert are reused when not found in the cluster and a
    client cert is requested
    """
    reset()
    session = setup_session(
        app_config=INTERNAL_TLS_OVERRIDES,
        deploy_manager=MockDeployManager(),
    )
    # Pre-populate the secret
    test_key, test_crt = set_tls_ca_secret(session)

    # Create the chart and render it
    def doit(session, scope):
        crt = tls_context.get_client_cert(session, scope, encode=False)
        assert crt is not None
        ca_key, ca_crt = get_internal_ctxt_ca_pair(session)
        assert crt == ca_crt
        assert crt == test_crt

    chart = make_stub_chart(session, doit)
    chart.to_config(session)


def test_reuse_when_found_get_server():
    """Test that the CA key/cert are reused when not found in the cluster and a
    server key/cert pair is requested
    """
    reset()
    session = setup_session(
        app_config=INTERNAL_TLS_OVERRIDES,
        deploy_manager=MockDeployManager(),
    )
    # Pre-populate the secret
    test_key, test_crt = set_tls_ca_secret(session)

    # Create the chart and render it
    def doit(session, scope):
        tls_context.request_server_key_cert_pair(session, scope, [])
        key, crt = tls_context.get_server_key_cert_pair(session, scope, encode=False)
        assert key is not None
        assert crt is not None
        ca_key, ca_crt = get_internal_ctxt_ca_pair(session)
        assert crt != ca_crt
        assert crt != test_crt
        assert key != ca_key
        assert key != test_key

    chart = make_stub_chart(session, doit)
    chart.to_config(session)


def test_state_shared_between_components():
    """Test two independent components can share the same state"""

    # Create the chart
    crts = set()

    def doit(session, scope):
        crt = tls_context.get_client_cert(session, scope, encode=False)
        assert crt is not None
        crts.add(crt)

    # Trigger the render
    reset()
    session = setup_session(
        app_config=INTERNAL_TLS_OVERRIDES,
    )
    chart1 = make_stub_chart(session, doit, name="chart1")
    chart2 = make_stub_chart(session, doit, name="chart2")
    chart1.to_config(session)
    chart2.to_config(session)
    assert len(crts) == 1


## Get Functions ###############################################################
#
# These tests exercise the functionality to get client and server TLS data
##


def test_encode_flag_server():
    """Test that the encode flag properly performs base64 encoding only if set
    to True for get_server_key_cert_pair
    """

    def doit(session, scope):
        tls_context.request_server_key_cert_pair(session, scope, [])

        key_enc, crt_enc = tls_context.get_server_key_cert_pair(
            session, scope, encode=True
        )
        key, crt = tls_context.get_server_key_cert_pair(session, scope, encode=False)
        common.b64_secret(key) == key_enc
        common.b64_secret(crt) == crt_enc

    # Trigger the render
    reset()
    session = setup_session(
        app_config=INTERNAL_TLS_OVERRIDES,
    )
    chart = make_stub_chart(session, doit)
    chart.to_config(session)


def test_encode_flag_client():
    """Test that the encode flag properly performs base64 encoding only if set
    to True for get_client_cert
    """

    def doit(session, scope):
        crt_enc = tls_context.get_client_cert(session, scope, encode=True)
        crt = tls_context.get_client_cert(session, scope, encode=False)
        common.b64_secret(crt) == crt_enc

    # Trigger the render
    reset()
    session = setup_session(
        app_config=INTERNAL_TLS_OVERRIDES,
    )
    chart = make_stub_chart(session, doit)
    chart.to_config(session)


## Server Communication ########################################################
#
# These tests validate that the communication paradigms work as expected
##


class ServerProcess(multiprocessing.Process):
    """This is a thread implementation which will run a simple Flask server and
    handle get reqeusts in an independent thread.
    """

    HELLO_WORLD = "Hello World!"

    def __init__(self, port, key_file, crt_file, ca_crt_file):
        self.port = port
        self.key_file = key_file
        self.crt_file = crt_file
        self.ca_crt_file = ca_crt_file
        self.started = multiprocessing.Event()
        super().__init__(target=self._start_server)

    def _start_server(self):
        app = Flask("stub")

        @app.route("/")
        def hello():
            return self.HELLO_WORLD

        self.started.set()
        app.run(port=self.port, ssl_context=(self.crt_file, self.key_file))

    def wait_for_boot(self):
        self.started.wait()
        time.sleep(0.1)

    def shutdown(self):
        self.terminate()


@contextmanager
def maybe_temp_dir():
    working_dir = os.environ.get("WORKING_DIR")
    if working_dir is not None:
        os.makedirs(working_dir, exist_ok=True)
        yield working_dir
    else:
        with tempfile.TemporaryDirectory() as working_dir:
            yield working_dir


# Use a stub chart to generate the pem content
@contextmanager
def gen_server_setup(session, server_number):
    generated = {}

    def doit(session, scope):
        tls_context.request_server_key_cert_pair(
            session, scope, ["localhost"], ["127.0.0.1"]
        )
        generated["ca_crt_pem"] = tls_context.get_client_cert(
            session, scope, encode=False
        )
        (
            generated["server_key_pem"],
            generated["server_crt_pem"],
        ) = tls_context.get_server_key_cert_pair(session, scope, encode=False)

    chart = make_stub_chart(session, doit, name=f"server-{server_number}")
    chart.to_config(session)

    # Save the files to the working dir
    with maybe_temp_dir() as working_dir:
        ca_crt_file = os.path.join(working_dir, f"ca_crt{server_number}.pem")
        with open(ca_crt_file, "w") as f:
            f.write(generated["ca_crt_pem"])
        server_key_file = os.path.join(working_dir, f"server_key{server_number}.pem")
        with open(server_key_file, "w") as f:
            f.write(generated["server_key_pem"])
        server_crt_file = os.path.join(working_dir, f"server_crt{server_number}.pem")
        with open(server_crt_file, "w") as f:
            f.write(generated["server_crt_pem"])

        # Also save out the ca key for debugging
        ca_key, _ = get_tls_context(session)._component.get_ca_key_cert()
        ca_key_file = os.path.join(working_dir, f"ca_key{server_number}.pem")
        with open(ca_key_file, "w") as f:
            f.write(ca_key)

        yield ca_crt_file, server_key_file, server_crt_file


def test_client_server_verify():
    """Test that a server with a generated key/cert pair and a client using the
    shared client cert can successfully communicate over TLS with verification
    enabled
    """

    # Generate the setup for the server
    session = setup_session(app_config=INTERNAL_TLS_OVERRIDES)
    with gen_server_setup(session, 1) as (
        ca_crt_file,
        server_key_file,
        server_crt_file,
    ):

        # Launch a server that will use the server pair
        port = get_available_port()
        server = ServerProcess(
            port=port,
            key_file=server_key_file,
            crt_file=server_crt_file,
            ca_crt_file=ca_crt_file,
        )
        server.start()
        server.wait_for_boot()

        # Make a request using the CA cert to verify
        res = requests.get(f"https://localhost:{port}", verify=ca_crt_file)
        res.raise_for_status()
        assert res.text == ServerProcess.HELLO_WORLD

        # Shut the server down
        server.shutdown()
        server.join()


def test_client_multi_server_verify():
    """Test that a client using the common cert can use it to communicate with
    multiple servers serving different key/cert pairs
    """

    session = setup_session(
        app_config=INTERNAL_TLS_OVERRIDES,
    )
    # Generate the setup for two servers
    with gen_server_setup(session, 1) as (
        ca_crt_file1,
        server_key_file1,
        server_crt_file1,
    ):
        with gen_server_setup(session, 2) as (
            ca_crt_file2,
            server_key_file2,
            server_crt_file2,
        ):

            # Make sure the two CA certs are the same
            with open(ca_crt_file1, "r") as f:
                ca1 = f.read()
            with open(ca_crt_file2, "r") as f:
                ca2 = f.read()
            assert ca1 == ca2
            ca_crt_file = ca_crt_file1

            # Launch two servers with the different key/cert pairs
            port1 = get_available_port()
            port2 = get_available_port()
            server1 = ServerProcess(
                port=port1,
                key_file=server_key_file1,
                crt_file=server_crt_file1,
                ca_crt_file=ca_crt_file1,
            )
            server2 = ServerProcess(
                port=port2,
                key_file=server_key_file2,
                crt_file=server_crt_file2,
                ca_crt_file=ca_crt_file2,
            )
            server1.start()
            server2.start()
            server1.wait_for_boot()
            server2.wait_for_boot()

            # Make a request to both servers using the same CA cert
            res = requests.get(f"https://localhost:{port1}", verify=ca_crt_file)
            res.raise_for_status()
            assert res.text == ServerProcess.HELLO_WORLD
            res = requests.get(f"https://localhost:{port2}", verify=ca_crt_file)
            res.raise_for_status()
            assert res.text == ServerProcess.HELLO_WORLD

            # Shut the server down
            server1.shutdown()
            server2.shutdown()
            server1.join()
            server2.join()


def test_get_tls_context_config_overrides():
    """Test that giving config overrides to get_tls_context takes precedende
    over backend config
    """
    session = setup_session(app_config={"tls": {"type": "invalid"}})
    get_tls_context(session, config_overrides=INTERNAL_TLS_OVERRIDES["tls"])


def test_reregister_ok():
    """Make sure that a type can be re-registered without raising. This is
    needed when registration is done in a derived library that uses the PWM and
    therefore re-imports the derived implementation.
    """
    register_tls_context_type(InternalTlsContext)



"""
Tests for the library config module

NOTE: Python makes it hard to change env vars in a way that will effect import
    time, so we're relying on the fact that aconfig is well tested and not
    actually validating the env-var override behavior!
"""

# Third Party
import pytest

# First Party
import aconfig

# Local
from oper8 import config


def test_config_keys():
    """Make sure that at least one expected key is present"""
    assert hasattr(config, "deploy_retries")
    assert isinstance(config.deploy_retries, int)


########################
## get_invalid_params ##
########################


def test_get_invalid_params_all_valid_params():
    """Test that get_invalid_params returns no invalid params when all are set
    to valid values
    """
    assert not config.validation.get_invalid_params(
        config=aconfig.Config({"key": 1}),
        validation_config=aconfig.Config({"key": {"type": "int", "min": 0, "max": 1}}),
    )


def test_get_invalid_params_all_invalid_params():
    """Test that get_invalid_params returns an invalid param when all are set
    to invalid values
    """
    assert config.validation.get_invalid_params(
        config=aconfig.Config({"key": 3}),
        validation_config=aconfig.Config(
            {"key": {"type": "int", "min": 0, "max": 1}},
        ),
    ) == ["key"]


def test_get_invalid_params_some_invalid_params():
    """Test that get_invalid_params returns only the invalid parameters when
    some are invalid and some are valid
    """
    assert config.validation.get_invalid_params(
        config=aconfig.Config({"key": 3, "str": "foo"}),
        validation_config=aconfig.Config(
            {
                "key": {"type": "int", "min": 0, "max": 1},
                "str": {"type": "str", "min_len": 1},
            },
        ),
    ) == ["key"]


#####################
## parameter types ##
#####################


def test_number_parameter():
    """Test all validation cases for _NumberParameter"""
    ParamType = config.validation._NumberParameter

    # Valid Cases
    assert ParamType().validate(1)
    assert ParamType().validate(1.2)
    assert ParamType(min=0).validate(1)
    assert ParamType(max=1).validate(0.5)
    assert ParamType(optional=True).validate(None)

    # Invalid Cases
    assert not ParamType().validate("not a number")
    assert not ParamType(min=0).validate(-1)
    assert not ParamType(max=1).validate(1.5)
    assert not ParamType(optional=False).validate(None)


def test_int_parameter():
    """Test all validation cases for _IntParameter"""
    ParamType = config.validation._IntParameter

    # Valid Cases
    assert ParamType().validate(1)
    assert ParamType(min=0).validate(1)
    assert ParamType(max=1).validate(1)
    assert ParamType(optional=True).validate(None)

    # Invalid Cases
    assert not ParamType().validate("not an int")
    assert not ParamType().validate(1.2)
    assert not ParamType(min=0).validate(-1)
    assert not ParamType(max=1).validate(2)
    assert not ParamType(optional=False).validate(None)


def test_float_parameter():
    """Test all validation cases for _FloatParameter"""
    ParamType = config.validation._FloatParameter

    # Valid Cases
    assert ParamType().validate(1.0)
    assert ParamType(min=0.0).validate(1.0)
    assert ParamType(max=1.0).validate(1.0)
    assert ParamType(optional=True).validate(None)

    # Invalid Cases
    assert not ParamType().validate("not an int")
    assert not ParamType().validate(1)
    assert not ParamType(min=0).validate(-1.0)
    assert not ParamType(max=1).validate(2.0)
    assert not ParamType(optional=False).validate(None)


def test_str_parameter():
    """Test all validation cases for _StrParameter"""
    ParamType = config.validation._StrParameter

    # Valid Cases
    assert ParamType().validate("test")
    assert ParamType(min_len=1).validate("test")
    assert ParamType(max_len=4).validate("test")
    assert ParamType(optional=True).validate(None)

    # Invalid Cases
    assert not ParamType().validate(1)
    assert not ParamType().validate(b"test")
    assert not ParamType(min_len=1).validate("")
    assert not ParamType(max_len=3).validate("test")
    assert not ParamType(optional=False).validate(None)


def test_bool_parameter():
    """Test all validation cases for _BoolParameter"""
    ParamType = config.validation._BoolParameter

    # Valid Cases
    assert ParamType().validate(True)
    assert ParamType().validate(False)
    assert ParamType(optional=True).validate(None)

    # Invalid Cases
    assert not ParamType().validate(1)
    assert not ParamType(optional=False).validate(None)


def test_enum_parameter():
    """Test all validation cases for _EnumParameter"""
    ParamType = config.validation._EnumParameter

    # Invalid Construction
    with pytest.raises(AssertionError):
        ParamType(values=[])
    with pytest.raises(AssertionError):
        ParamType(values="test")

    # Valid Cases
    assert ParamType(values=[1, "two", None]).validate(1)
    assert ParamType(values=[1, "two", None]).validate("two")
    assert ParamType(values=[1, "two", None]).validate(None)
    assert ParamType(values=[1, "two"], optional=True).validate(None)

    # Invalid Cases
    assert not ParamType(values=[1, "two", None]).validate(2)
    assert not ParamType(values=[1, "two"], optional=False).validate(None)


def test_list_parameter():
    """Test all validation cases for _ListParameter"""
    ParamType = config.validation._ListParameter

    # Valid Cases
    assert ParamType().validate([])
    assert ParamType().validate([1.2])
    assert ParamType(min_len=0).validate([1, "two"])
    assert ParamType(max_len=1).validate([0.5])
    assert ParamType(item_type="int").validate([1, 2, 3])
    assert ParamType(optional=True).validate(None)

    # Invalid Cases
    assert not ParamType().validate("not a list")
    assert not ParamType(min_len=1).validate([])
    assert not ParamType(max_len=1).validate([1, 2])
    assert not ParamType(item_type="str").validate([1, 2])
    assert not ParamType(item_type="str").validate(["one", 2])
    assert not ParamType(optional=False).validate(None)


#############
## factory ##
#############


def test_construct_parameter_all_known_types():
    """Make sure that all known types (as of the time of this writing) can be
    constructed via the factory
    """
    assert isinstance(
        config.validation._construct_parameter({"type": "number"}),
        config.validation._NumberParameter,
    )
    assert isinstance(
        config.validation._construct_parameter({"type": "int", "min": 1}),
        config.validation._IntParameter,
    )
    assert isinstance(
        config.validation._construct_parameter({"type": "float", "max": 2}),
        config.validation._FloatParameter,
    )
    assert isinstance(
        config.validation._construct_parameter({"type": "str", "min_len": 1}),
        config.validation._StrParameter,
    )
    assert isinstance(
        config.validation._construct_parameter({"type": "bool"}),
        config.validation._BoolParameter,
    )
    assert isinstance(
        config.validation._construct_parameter({"type": "enum", "values": [1]}),
        config.validation._EnumParameter,
    )


def test_construct_parameter_extra_params_error():
    """Make sure that a param construction call with bad arguments raises an
    error
    """
    with pytest.raises(TypeError):
        config.validation._construct_parameter({"type": "number", "foo": "bar"})


def test_construct_parameter_missing_params_error():
    """Make sure that a param construction call that is missing required params
    raises an error
    """
    with pytest.raises(TypeError):
        config.validation._construct_parameter({"type": "enum"})


def test_construct_parameter_unknown_type():
    """Test that constructing a parameter returns None when the value of 'type'
    doesn't match a configured type
    """
    assert config.validation._construct_parameter({"type": "foobar"}) is None


#############
## parsing ##
#############


def test_parse_validation_config_nested_key():
    """Make sure that parsing a validation config with nested keys works"""
    assert list(
        config.validation._parse_validation_config(
            aconfig.Config({"foo": {"bar": {"baz": {"type": "int"}}}})
        ).keys()
    ) == ["foo.bar.baz"]


def test_parse_validation_config_nested_type_key():
    """Make sure that parsing a validation config is robust to having the key
    'type' not represent an actual parameter
    """
    assert list(
        config.validation._parse_validation_config(
            aconfig.Config({"foo": {"type": {"baz": {"type": "int"}}}})
        ).keys()
    ) == ["foo.type.baz"]



"""
Module for the ThreadBase Class
"""

# Standard
import threading

# First Party
import alog

# Local
from ....deploy_manager import DeployManagerBase
from ..leader_election import DryRunLeadershipManager, LeadershipManagerBase

log = alog.use_channel("TRDUTLS")


class ThreadBase(threading.Thread):
    """Base class for all other thread classes. This class handles generic starting, stopping,
    and leadership functions"""

    def __init__(
        self,
        name: str = None,
        daemon: bool = None,
        deploy_manager: DeployManagerBase = None,
        leadership_manager: LeadershipManagerBase = None,
    ):
        """Initialize class and store required instance variables. This function
        is normally overriden by subclasses that pass in static name/daemon variables

        Args:
            name:str=None
                The name of the thread to manager
            daemon:bool=None
                Whether python should wait for this thread to stop before exiting
            deploy_manager: DeployManagerBase = None
                The deploy manager available to this thread during start()
            leadership_manager: LeadershipManagerBase = None
                The leadership_manager for tracking elections
        """
        self.deploy_manager = deploy_manager
        self.leadership_manager = leadership_manager or DryRunLeadershipManager()
        self.shutdown = threading.Event()
        super().__init__(name=name, daemon=daemon)

    ## Abstract Interface ######################################################
    #
    # These functions must be implemented by child classes
    ##
    def run(self):
        """Control loop for the thread. Once this function exits the thread stops"""
        raise NotImplementedError()

    ## Base Class Interface ####################################################
    #
    # These methods MAY be implemented by children, but contain default
    # implementations that are appropriate for simple cases.
    #
    ##

    def start_thread(self):
        """If the thread is not already alive start it"""
        if not self.is_alive():
            log.info("Starting %s: %s", self.__class__.__name__, self.name)
            self.start()

    def stop_thread(self):
        """Set the shutdown event"""
        log.info("Stopping %s: %s", self.__class__.__name__, self.name)
        self.shutdown.set()

    def should_stop(self) -> bool:
        """Helper to determine if a thread should shutdown"""
        return self.shutdown.is_set()

    def check_preconditions(self) -> bool:
        """Helper function to check if the thread should shutdown or reacquire leadership"""
        if self.should_stop():
            return False

        if self.leadership_manager and not self.leadership_manager.is_leader():
            log.debug3("Waiting for leadership")
            self.leadership_manager.acquire()

        return True



"""
Tests for the WatchManagerBase base class
"""

# Standard
import threading
import time

# Third Party
import pytest

# Local
from oper8.test_helpers.helpers import DummyController
from oper8.watch_manager.base import WatchManagerBase

## Helpers #####################################################################


class DummyWatchManager(WatchManagerBase):
    def __init__(
        self,
        controller_type,
        watch_success=True,
        stop_wait=0.0,
    ):
        super().__init__(controller_type)
        self.watching = False
        self.watch_success = watch_success
        self.stop_wait = stop_wait

    def watch(self):
        if self.watch_success:
            self.watching = True
            return True
        return False

    def wait(self):
        while self.watching:
            time.sleep(0.05)

    def stop(self):
        if self.stop_wait:
            threading.Thread(target=self._delayed_stop).start()
        else:
            self.watching = False

    def _delayed_stop(self):
        time.sleep(self.stop_wait)
        self.watching = False


class DummyController2(DummyController):
    group = "asdf.qwer"
    version = "v1"
    kind = "Widget"


@pytest.fixture(autouse=True)
def reset_globals():
    """This helper is only used in tests to "reset" the state of the global
    watches dict
    """
    WatchManagerBase._ALL_WATCHES = {}


## Tests #######################################################################


def test_constructor_properties():
    """Test that the base class properties are set on the watch manager"""
    wm = DummyWatchManager(DummyController)
    assert wm.controller_type == DummyController
    assert wm.group == DummyController.group
    assert wm.version == DummyController.version
    assert wm.kind == DummyController.kind


def test_constructor_registrations():
    """Test that all constructed watch managers get registered"""
    wm1 = DummyWatchManager(DummyController)
    wm2 = DummyWatchManager(DummyController2)
    assert len(WatchManagerBase._ALL_WATCHES) == 2
    assert str(wm1) in WatchManagerBase._ALL_WATCHES
    assert str(wm2) in WatchManagerBase._ALL_WATCHES


def test_constructor_no_duplicate_watches():
    """Test that all constructed watch managers get registered"""
    DummyWatchManager(DummyController)
    with pytest.raises(AssertionError):
        DummyWatchManager(DummyController)


def test_start_stop_all_blocking():
    """Test that calling start_all and stop_all with block set to true do indeed
    block and correctly start/stop all managers
    """
    wm1 = DummyWatchManager(DummyController)
    wm2 = DummyWatchManager(DummyController2)

    # Run start_all in a thread so that we can stop it
    thrd = threading.Thread(target=WatchManagerBase.start_all)
    thrd.start()
    time.sleep(0.1)

    # Make sure both watch managers are watching
    assert wm1.watching
    assert wm2.watching

    # Make sure the thread is blocked
    assert thrd.is_alive()

    # Stop all watch managers, and make sure all watch managers are not watching
    WatchManagerBase.stop_all()
    assert not wm1.watching
    assert not wm2.watching


def test_start_all_blocking_failure():
    """Test that calling start_all when one of the managers fails to start
    cleanly shuts down any started managers
    """
    # NOTE: failure is on wm1 because it comes second alphabetically
    wm1 = DummyWatchManager(DummyController, watch_success=False)
    wm2 = DummyWatchManager(DummyController2)

    # Start them and make sure it returns failure
    assert not WatchManagerBase.start_all()
    assert not wm1.watching
    assert not wm2.watching



"""
Ansible-based implementation of the WatchManager
"""

# Standard
from typing import Optional, Type
import copy
import json
import os
import shlex
import subprocess
import time

# Third Party
import yaml

# First Party
import alog

# Local
from ... import config
from ...controller import Controller
from ...exceptions import assert_config
from ..base import WatchManagerBase

log = alog.use_channel("ANSBL")

DEFAULT_ENTRYPOINT = "/usr/local/bin/ansible-operator run --watches-file=./watches.yaml"


class AnsibleWatchManager(WatchManagerBase):
    """The AnsibleWatchManager uses the core of an ansible-based operator to
    manage watching resources. The key elements are:

    1. Manage a `watches.yaml` file for all watched resources
    2. Manage a playbook for each watched resource
    3. Manage the ansible operator's executable as a subprocess
    """

    # Shared singleton process used to manage all watches via ansible
    ANSIBLE_PROCESS = None

    # Defaults for initialization args held separately to allow for override
    # precedence order
    _DEFAULT_INIT_KWARGS = {
        "ansible_base_path": "/opt/ansible",
        "ansible_entrypoint": DEFAULT_ENTRYPOINT,
        "ansible_args": "",
        "manage_status": False,
        "watch_dependent_resources": False,
        "reconcile_period": "10m",
        "playbook_parameters": None,
    }

    def __init__(
        self,
        controller_type: Type[Controller],
        *,
        ansible_base_path: Optional[str] = None,
        ansible_entrypoint: Optional[str] = None,
        ansible_args: Optional[str] = None,
        manage_status: Optional[bool] = None,
        watch_dependent_resources: Optional[bool] = None,
        reconcile_period: Optional[str] = None,
        playbook_parameters: Optional[dict] = None,
    ):
        """Construct with the core watch binding and configuration args for the
        watches.yaml and playbook.yaml files.

        NOTE: All args may be overridden in the `ansible_watch_manager` section
            of the library config. The precedence order is:

        1. Directly passed arguments
        2. Config values
        3. Default values from code

        A passed None value in any of these is considered "unset"

        Args:
            controller_type:  Type[Controller],
                The Controller type that will manage this group/version/kind

        Kwargs:
            ansible_base_path:  str
                The base path where the ansible runtime will be run. This is
                also used to determine where the watches.yaml and playbooks will
                be managed.
            ansible_entrypoint:  str
                The command to use to run ansible
            ansible_args: str
                Additional flags to be passed to `ansible_entrypoint`
            manage_status:  bool
                Whether or not to let ansible manage status on the CR
            watch_dependent_resources:  bool
                Whether or not to trigger a reconciliation on change to
                dependent resources.
            reconcile_period:  str
                String representation of the time duration to use for periodic
                reconciliations
            playbook_parameters:  dict
                Parameters to use to configure the k8s_application module in the
                playbook
        """
        # Make sure that the shared ansible process is not already started
        assert (
            self.ANSIBLE_PROCESS is None
        ), "Cannot create an AnsibleWatchManager after starting another AnsibleWatchManager"

        # Set up the function arguments based on override precedence
        ansible_base_path = self._init_arg("ansible_base_path", ansible_base_path, str)
        ansible_entrypoint = self._init_arg(
            "ansible_entrypoint", ansible_entrypoint, str
        )
        ansible_args = self._init_arg("ansible_args", ansible_args, str)
        manage_status = self._init_arg("manage_status", manage_status, bool)
        watch_dependent_resources = self._init_arg(
            "watch_dependent_resources", watch_dependent_resources, bool
        )
        reconcile_period = self._init_arg("reconcile_period", reconcile_period, str)
        playbook_parameters = self._init_arg(
            "playbook_parameters", playbook_parameters, dict
        )

        super().__init__(controller_type)
        self._ansible_base_path = ansible_base_path
        self._ansible_entrypoint = ansible_entrypoint
        self._ansible_args = ansible_args

        # Create the playbook
        playbook_path = self._add_playbook(playbook_parameters)

        # Create the entry in the watches.yaml
        self._add_watch_entry(
            playbook_path=playbook_path,
            manage_status=manage_status,
            watch_dependent_resources=watch_dependent_resources,
            reconcile_period=reconcile_period,
            add_finalizer=controller_type.has_finalizer,
            disable_vcs=getattr(controller_type, "disable_vcs", None),
        )

    ## Interface ###############################################################

    def watch(self) -> bool:
        """Start the global ansible process if not already started

        NOTE: This is intentionally not thread safe! The watches should all be
            managed from the primary entrypoint thread.

        Returns:
            success:  bool
                True if the asible process is running correctly
        """
        cls = self.__class__
        if cls.ANSIBLE_PROCESS is None:
            log.info("Starting ansible watch process")
            env = copy.deepcopy(os.environ)
            env["ANSIBLE_LIBRARY"] = self._ansible_library_path()
            env["ANSIBLE_ROLES_PATH"] = self._ansible_roles_path()
            cls.ANSIBLE_PROCESS = (
                subprocess.Popen(  # pylint: disable=consider-using-with
                    shlex.split(
                        " ".join((self._ansible_entrypoint, self._ansible_args)).strip()
                    ),
                    cwd=self._ansible_base_path,
                    env=env,
                )
            )

        # If the process does not have a returncode on poll, it's up. This is a
        # point-in-time statement. There's no way for this code to actually
        # validate the state of the process since it may crash at any
        # indeterminate time after starting.
        return self.ANSIBLE_PROCESS.poll() is None

    def wait(self):
        """Wait for the ansible process to terminate"""
        if self.ANSIBLE_PROCESS is not None:
            self.ANSIBLE_PROCESS.wait()

    def stop(self):
        """Attempt to terminate the ansible process. This asserts that the
        process has been created in order to avoid race conditions with a None
        check.
        """
        assert self.ANSIBLE_PROCESS is not None, "Cannot stop before watching"
        log.info("Killing shared ansible process")
        self.ANSIBLE_PROCESS.terminate()
        kill_start_time = time.time()
        while (
            self.ANSIBLE_PROCESS.poll() is None
            and time.time() - kill_start_time
            < config.ansible_watch_manager.kill_max_wait
        ):
            time.sleep(0.001)
        assert (
            self.ANSIBLE_PROCESS.poll() is not None
        ), "The only way to shut down ansible is with a sledge hammer!"

    ## Implementation Details ##################################################

    @classmethod
    def _init_arg(cls, arg_name, passed_value, arg_type):
        """Helper to enforce init arg precedence"""
        if passed_value is not None:
            return passed_value
        config_value = config.ansible_watch_manager.get(arg_name)
        if config_value is not None:
            if arg_type is not None and not isinstance(config_value, arg_type):
                assert_config(
                    isinstance(config_value, str),
                    f"Invalid type for ansible_watch_manager.{arg_name}: "
                    + "{type(config_value)} should be {arg_type}",
                )
                if arg_type is bool:
                    config_value = config_value.lower() == "true"
                elif arg_type is dict:
                    config_value = json.loads(config_value)
                assert_config(
                    isinstance(config_value, arg_type),
                    f"Cannot convert ansible_watch_manager.{arg_name} from str to {arg_type}",
                )
            return config_value
        assert (
            arg_name in cls._DEFAULT_INIT_KWARGS
        ), f"Programming Error: Unsupported init kwarg: {arg_name}"
        return cls._DEFAULT_INIT_KWARGS[arg_name]

    def _add_playbook(self, playbook_parameters):
        """Create a playbook for this watch"""

        # Open the base template for playbooks
        playbook_base_path = os.path.join(
            self._resources_path(),
            "playbook-base.yaml",
        )
        with open(playbook_base_path, encoding="utf-8") as handle:
            playbook_base = yaml.safe_load(handle)

        # Add the provided variables
        module_vars = playbook_parameters or {}
        module_vars.setdefault("strict_versioning", False)
        kind = self.controller_type.kind.lower()
        log_file = f"{kind}.{{{{ ansible_operator_meta.name }}}}.log"
        log_dir = config.ansible_watch_manager.log_file_dir
        if log_dir is not None:
            log.debug2("Adding log dir: %s", log_dir)
            log_file = os.path.join(log_dir, log_file)
        module_vars.setdefault("log_file", log_file)
        playbook_base[0]["tasks"][0]["vars"] = module_vars

        # Add the controller_class
        controller_class = (
            f"{self.controller_type.__module__}.{self.controller_type.__name__}"
        )
        log.debug3("controller_class: %s", controller_class)
        module_vars["controller_class"] = controller_class

        # Add the full_cr template
        group_template = self.group.lower().replace(".", "_").replace("-", "_")
        cr_template = f"{{{{ _{group_template}_{self.kind.lower()} }}}}"
        module_vars["full_cr"] = cr_template

        # Write it out to the right place
        log.debug3(
            "%s/%s/%s playbook vars: %s",
            self.group,
            self.version,
            self.kind,
            module_vars,
        )
        playbook_path = os.path.join(
            self._ansible_base_path, f"playbook-{self.kind.lower()}.yaml"
        )
        with open(playbook_path, "w", encoding="utf-8") as handle:
            yaml.dump(playbook_base, handle)
        return playbook_path

    def _add_watch_entry(  # pylint: disable=too-many-arguments
        self,
        playbook_path: str,
        manage_status: bool,
        watch_dependent_resources: bool,
        reconcile_period: str,
        add_finalizer: bool,
        disable_vcs: Optional[bool],
    ):
        """Add an entry to the watches.yaml file, creating it if needed"""

        # Load the current watches.yaml content, or start fresh
        watches_path = os.path.join(self._ansible_base_path, "watches.yaml")
        if os.path.exists(watches_path):
            with open(watches_path, encoding="utf-8") as handle:
                watches = yaml.safe_load(handle)
        else:
            watches = []

        # Make sure there is not already an entry for this watch
        matches = [
            (
                watch_entry["group"] == self.group
                and watch_entry["version"] == self.version
                and watch_entry["kind"] == self.kind
            )
            for watch_entry in watches
        ]
        assert True not in matches, (
            "Can't have multiple watch entries for the same group/version/kind! "
            + f"{self.group}/{self.version}/{self.kind}"
        )
        log.debug2("Adding new watch for %s", self)
        watch_entry = {
            "group": self.group,
            "version": self.version,
            "kind": self.kind,
            "vars": {"operation": "add"},
        }
        if disable_vcs is not None:
            str_val = str(not disable_vcs).lower()
            log.debug(
                "Adding watch variable [enable_ansible_vcs = '%s'] for %s/%s/%s",
                str_val,
                self.group,
                self.version,
                self.kind,
            )
            watch_entry["vars"]["enable_ansible_vcs"] = str_val
        watches.append(watch_entry)

        # Update the watch entry with the configuration for this watch
        watch_entry["playbook"] = playbook_path
        watch_entry["manageStatus"] = manage_status
        watch_entry["watchDependentResources"] = watch_dependent_resources
        watch_entry["reconcilePeriod"] = reconcile_period

        # If requested, add a version of the watch that manages the finalizer
        if add_finalizer:
            finalizer_name = self.controller_type.finalizer
            log.debug2("Adding finalizer: %s", finalizer_name)
            watch_entry["finalizer"] = {
                "name": finalizer_name,
                "vars": {"operation": "remove"},
            }

        # Write the watches.yaml file back out
        with open(watches_path, "w", encoding="utf-8") as handle:
            yaml.dump(watches, handle)

    @staticmethod
    def _resources_path():
        """Get the path to the static resources for ansible"""
        return os.path.realpath(os.path.join(os.path.dirname(__file__), "resources"))

    @staticmethod
    def _ansible_library_path():
        """Get the absolute path to the ansible library with the k8s_applicaiton
        module
        """
        return os.path.realpath(os.path.join(os.path.dirname(__file__), "modules"))

    @classmethod
    def _ansible_roles_path(cls):
        """Get the absolute path to the ansible roles"""
        return os.path.join(cls._resources_path(), "roles")



"""
Tests for the ReconcileProcessEntrypoint
"""
# Standard
from contextlib import contextmanager
from multiprocessing.connection import Connection
from unittest import mock
import logging
import multiprocessing
import time

# Third Party
import pytest

# First Party
import alog

# Local
from oper8 import DeployManagerBase
from oper8.deploy_manager import DeployMethod
from oper8.deploy_manager.dry_run_deploy_manager import DryRunDeployManager
from oper8.deploy_manager.kube_event import KubeEventType
from oper8.reconcile import ReconciliationResult
from oper8.test_helpers.helpers import (
    DummyController,
    configure_logging,
    library_config,
)
from oper8.test_helpers.pwm_helpers import make_managed_object
from oper8.watch_manager.python_watch_manager.filters.filters import DisableFilter
from oper8.watch_manager.python_watch_manager.filters.manager import FilterManager
from oper8.watch_manager.python_watch_manager.reconcile_process_entrypoint import (
    ReconcileProcessDeployManager,
    create_and_start_entrypoint,
)
from oper8.watch_manager.python_watch_manager.utils.types import (
    ReconcileRequest,
    WatchRequest,
)

## Helpers #####################################################################

# NOTE: We need to keep this configure_logging call here since this file will be
#   re-imported inside the subprocess and the parent process log config is not
#   automatically recreated
configure_logging()
log = alog.use_channel("TEST")

context = multiprocessing.get_context("spawn")


class EntrypointDummyController(DummyController):
    pwm_filters = {"Foo.v1": [DisableFilter]}

    def __init__(self, config_defaults=None, deploy_manager=None):
        super().__init__(
            components=[
                {
                    "name": "foo",
                    "api_objects": [("foo", {"kind": "Foo", "apiVersion": "v1"})],
                },
                {
                    "name": "bar",
                    "api_objects": [("bar", {"kind": "Bar", "apiVersion": "v2"})],
                    "upstreams": ["foo"],
                },
                {
                    "name": "baz",
                    "api_objects": [("baz", {"kind": "Baz", "apiVersion": "v3"})],
                    "disabled": True,
                },
            ],
            config_defaults=config_defaults,
            deploy_manager=deploy_manager,
        )


class ParentDummyController(DummyController):
    pwm_subsystems = [EntrypointDummyController]
    version = "v41"

    def __init__(self, config_defaults=None, deploy_manager=None):
        super().__init__(
            components=[
                {
                    "name": "foo",
                    "api_objects": [
                        ("foo", {"kind": "Foo", "apiVersion": "foo.bar.com/v42"})
                    ],
                },
            ],
            config_defaults=config_defaults,
            deploy_manager=deploy_manager,
        )


@contextmanager
def mock_entrypoint_deploy_manager():
    original_deploy = DryRunDeployManager._deploy

    def mocked_deploy(
        self,
        resource_definitions,
        call_watches=True,
        manage_owner_references=True,
        method=DeployMethod.DEFAULT,
    ):
        success, changes = original_deploy(
            self,
            resource_definitions=resource_definitions,
            call_watches=call_watches,
            manage_owner_references=manage_owner_references,
            method=method,
        )
        if not success or not changes:
            log.debug2("Failed to deploy resources")
            return success, changes

        DryRunDeployManager._apply_resource = (
            lambda self, resource_definition: resource_definition
        )
        for resource in resource_definitions:
            resource_metadata = resource.get("metadata")
            _, updated_resource = self.get_object_current_state(
                resource.get("kind"),
                resource_metadata.get("name"),
                resource_metadata.get("namespace"),
                resource.get("apiVersion"),
            )
            self._apply_resource(resource)

        return success, changes

    original_disable = DryRunDeployManager.disable

    def mocked_disable(self, resource_definitions):
        success, changes = original_disable(
            self, resource_definitions=resource_definitions
        )
        if not success or not changes:
            log.debug2("Failed to disable resources")
            return success, changes

        DryRunDeployManager._disable = lambda self, resource_definition: changes
        for resource in resource_definitions:
            self._disable(resource)

        return success, changes

    patched_object = mock.patch.object(
        ReconcileProcessDeployManager, "__bases__", (DryRunDeployManager,)
    )
    with patched_object:
        patched_object.is_local = True
        patched_object_deploy = mock.patch.object(
            ReconcileProcessDeployManager, "_deploy", mocked_deploy
        )
        with patched_object_deploy:
            patched_object_disable = mock.patch.object(
                ReconcileProcessDeployManager, "disable", mocked_disable
            )
            with patched_object_disable:
                yield


def mock_entrypoint_start(
    logging_queue: multiprocessing.Queue,
    request: ReconcileRequest,
    result_pipe: Connection,
    deploy_manager: DeployManagerBase,
    watch_dependent_resources=False,
    subsystem_rollout=False,
):
    with library_config(
        log_level="debug4",
        python_watch_manager={
            "watch_dependent_resources": watch_dependent_resources,
            "subsystem_rollout": subsystem_rollout,
        },
    ), mock_entrypoint_deploy_manager():
        create_and_start_entrypoint(
            logging_queue, request, result_pipe, deploy_manager=deploy_manager
        )


def flatten_multi_level_list(list_obj):
    if list_obj == []:
        return list_obj
    if isinstance(list_obj[0], list):
        return flatten_multi_level_list(list_obj[0]) + flatten_multi_level_list(
            list_obj[1:]
        )
    return list_obj[:1] + flatten_multi_level_list(list_obj[1:])


@pytest.mark.timeout(5)
def test_entrypoint_happy_path():
    recv_pipe, send_pipe = context.Pipe()
    resource = make_managed_object(api_version="foo.bar.com/v42")
    with mock_entrypoint_deploy_manager():
        dm = ReconcileProcessDeployManager(
            EntrypointDummyController, resource.definition, send_pipe
        )
        dm.deploy([resource.definition])

        logging_queue = context.Queue()
        request = ReconcileRequest(
            controller_type=EntrypointDummyController,
            type=KubeEventType.ADDED,
            resource=resource,
        )

        process = context.Process(
            target=mock_entrypoint_start, args=[logging_queue, request, send_pipe, dm]
        )
        process.start()

        time.sleep(2)
        process.join()
        assert process.exitcode == 0
        assert isinstance(logging_queue.get(), logging.LogRecord)
        assert isinstance(recv_pipe.recv(), ReconciliationResult)


@pytest.mark.timeout(5)
def test_entrypoint_subsystem():
    recv_pipe, send_pipe = context.Pipe()
    resource = make_managed_object(api_version="foo.bar.com/v41")
    with mock_entrypoint_deploy_manager():
        dm = ReconcileProcessDeployManager(
            ParentDummyController, resource.definition, send_pipe
        )
        dm.deploy([resource.definition])

        logging_queue = context.Queue()
        request = ReconcileRequest(
            controller_type=ParentDummyController,
            type=KubeEventType.ADDED,
            resource=resource,
        )

        process = context.Process(
            target=mock_entrypoint_start,
            args=[logging_queue, request, send_pipe, dm, False, True],
        )
        process.start()

        time.sleep(2)
        process.join()
        assert process.exitcode == 0
        assert isinstance(logging_queue.get(), logging.LogRecord)
        assert isinstance(recv_pipe.recv(), ReconciliationResult)


def test_entrypoint_dependent_watches():
    recv_pipe, send_pipe = context.Pipe()
    resource = make_managed_object(api_version="foo.bar.com/v42")
    with mock_entrypoint_deploy_manager():
        dm = ReconcileProcessDeployManager(
            EntrypointDummyController, resource.definition, send_pipe
        )
        dm.deploy([resource.definition])

        logging_queue = context.Queue()
        request = ReconcileRequest(
            controller_type=EntrypointDummyController,
            type=KubeEventType.ADDED,
            resource=resource,
        )

        process = context.Process(
            target=mock_entrypoint_start,
            args=[logging_queue, request, send_pipe, dm, True],
        )
        process.start()

        time.sleep(2)
        process.join()
        assert process.exitcode == 0
        assert isinstance(logging_queue.get(), logging.LogRecord)
        result_messages = []
        while recv_pipe.poll():
            result_messages.append(recv_pipe.recv())

        assert ReconciliationResult in [type(result) for result in result_messages]
        assert WatchRequest in [type(result) for result in result_messages]
        for result in result_messages:
            if (
                isinstance(result, WatchRequest)
                and result.watched.kind == "Foo"
                and result.watched.api_version == "v1"
            ):
                parsed_filters = FilterManager.from_info(result.filters_info)
                used_filters = flatten_multi_level_list(parsed_filters)

                assert DisableFilter in used_filters



"""Module contains helpers for processing a group of filters"""
# Standard
from typing import Any, Callable, List, Optional, Tuple, Type, Union
import inspect
import operator

# First Party
import alog

# Local
from ....deploy_manager import KubeEventType
from ....managed_object import ManagedObject
from ..utils import ClassInfo
from .filters import Filter

log = alog.use_channel("PWMFLTMAN")


## Conditional filters


def AndFilter(*args):  # pylint: disable=invalid-name
    """An "And" Filter is just a list of filters"""
    return list(args)


def OrFilter(*args):  # pylint: disable=invalid-name
    """An "Or" Filter is just a tuple of filters"""
    return tuple(args)


class FilterManager(Filter):
    """The FilterManager class helps process conditional filters and groups of filters.
    Filters that in a list are "anded" together while Filters in a tuple are "ored".
    This class also contains helpers to recursively convert between ClassInfo and Filters.
    """

    def __init__(
        self,
        filters: Union[List[Type[Filter]], Tuple[Type[Filter]]],
        resource: ManagedObject,
    ):
        """Initialize all filters in the provided group

        Args:
            filters: Union[List[Type[Filter]], Tuple[Type[Filter]]]
                The filters to manage
            resource: ManagedObject
                The initial resource
        """
        self.filters = self.__recursive_filter_init(filters, resource)

    ### Public Interface

    def update_and_test(
        self, resource: ManagedObject, event: KubeEventType
    ) -> Optional[bool]:
        """Recursively update and test each filter"""
        return self.__recursive_update_and_test(self.filters, resource, event)

    def test(self, resource: ManagedObject, event: KubeEventType) -> Optional[bool]:
        """Recursively test each filter"""
        # test with test_only set to True so nothing is updated
        return self.__recursive_update_and_test(
            self.filters, resource, event, test_only=True
        )

    def update(self, resource: ManagedObject):
        """Update each filter recursively"""
        # Update with update_only set to True so no tests are ran
        self.__recursive_update_and_test(  # pylint: disable=redundant-keyword-arg
            self, self.filters, resource, None, update_only=True
        )

    @classmethod
    def to_info(cls, filters: Type[Filter]) -> Type[ClassInfo]:
        """Helper function to convert from filters to ClassInfos. This is used for pickling and IPC

        Args:
            filters: Type[Filter]
                The filters to convert

        Returns:
            class_info: Type[ClassInfo]
                The class info objects describing the filter
        """
        return cls.__recursive_filter_info(filters)

    @classmethod
    def from_info(cls, info: Type[ClassInfo]) -> Type[Filter]:
        """Helper function to convert from ClassInfos to a filter


        Args:
            class_info: Type[ClassInfo]
                The classinfos to convert back into filters

        Returns:
            filters: Type[Filter]
                The converted filter objects
        """
        return cls.__recursive_filter_info(info)

    ### Private Helper Functions
    @classmethod
    def __recursive_filter_info(
        cls, descriptor: Union[Type[Filter], Type[ClassInfo]]
    ) -> Union[Type[Filter], Type[ClassInfo]]:
        """Recursive helper to convert from filters to class infos and back

        Args:
            descriptor: Union[Type[Filter],Type[ClassInfo]]
                Either the filter or class_info to convert

        Returns:
            type: Union[Type[Filter],Type[ClassInfo]]
                The converted types
        """

        def convert_filter_type(descriptor):
            """Generic function to convert between types"""

            # If we get a filter than we're converting to ClassInfo else
            # we're converting back to Filters
            if inspect.isclass(descriptor) and issubclass(descriptor, Filter):
                return ClassInfo.from_type(descriptor)
            if isinstance(descriptor, ClassInfo):
                return descriptor.to_class()
            # The instance must be a list or a tuple to be processed
            raise ValueError(
                f"Unknown type: {type(descriptor)} {descriptor} passed to convert_filter_type"
            )

        return cls.__recursive_map(descriptor, convert_filter_type)

    def __recursive_filter_init(
        self,
        filters: Union[List[Type[Filter]], Tuple[Type[Filter]], Type[Filter]],
        resource: ManagedObject,
    ) -> Union[List[Filter], Tuple[Filter], Filter]:
        """Helper function to recursively init each filter

        Args:
            filters: Union[List[Type[Filter]], Tuple[Type[Filter]], Type[Filter]]
                The filters to be initialized
            resource: ManagedObject
                The resource to pass to the filters
        Returns:
            filters: Union[List[Filter], Tuple[Filter], Filter]
                The initalized filters
        """

        def init_filter(filter_type: Type[Filter]) -> Filter:
            if not (inspect.isclass(filter_type) and issubclass(filter_type, Filter)):
                raise ValueError(
                    f"Unknown type: {type(filter_type)} passed to init_filter"
                )

            return filter_type(resource)

        return self.__recursive_map(filters, init_filter)

    def __recursive_update_and_test(  # pylint: disable=too-many-arguments, inconsistent-return-statements
        self,
        filters: Union[list, tuple, Filter],
        resource: ManagedObject,
        event: KubeEventType,
        update_only: bool = False,
        test_only: bool = False,
    ) -> Optional[bool]:
        """Helper function to recursively update, test, or both.

        Args:
            filters: Union[list, tuple, Filter]
                The current filters being tested. This is updated when recurring
            resource: ManagedObject
                The current resource being updated/tested
            event: KubeEventType,
                The current event type being updated/tested
            update_only: bool = False
                Whether to only update the filters
            test_only: bool = False,
                Whether to only test the filters

        Returns:
            result: Optional[bool]
                The result of the tests if it was ran
        """
        if update_only and test_only:
            raise ValueError("update_only and test_only can not both be True")

        # Check Initial object types and exit condition
        if isinstance(filters, Filter):
            # If instance is a filter than call either update or update_and_test
            # depending on the failed status
            if update_only:
                filters.update(resource)
                return
            if test_only:
                return filters.test(resource, event)

            return filters.update_and_test(resource, event)

        # If filter list is empty then immediately return success
        if not filters:
            return True

        return_value = None
        operation = operator.and_ if isinstance(filters, list) else operator.or_

        for filter_combo in filters:
            # Recursively processes the filter combo
            result = self.__recursive_update_and_test(
                filter_combo, resource, event, update_only, test_only
            )

            # If return_value has already been set then combine it with the most recent result
            # via the operation
            if result is not None:
                if return_value is not None:
                    return_value = operation(return_value, result)
                else:
                    return_value = result

            # There are two scenarios when filters only need to get updated not tested. The first
            # is when an "and" condition fails or when an "or" succeeds. In both instances we
            # know the end result so testing can be skipped for performance
            if (
                (not update_only and not test_only)
                and result
                and (
                    (operation == operator.and_ and not result)
                    or (operation == operator.or_ and result)
                )
            ):
                update_only = True

        # If no filter cared about the event then don't
        # reconcile
        if return_value is None:
            return False

        return return_value

    @classmethod
    def __recursive_map(
        cls, filters: Union[List[Any], Tuple[Any]], operation: Callable[[Filter], Any]
    ):
        """Helper function to map an operation onto every object in a filter chain

        Args:
            filters: Union[List[Any], Tuple[Any]]
                The filters to map onto
            op: Callable[[Filter],None]
                The function to map onto each filter
        """

        # Directly check tuple to ignore NamedTuples and subclasses
        if not (isinstance(filters, list) or type(filters) is tuple):
            return operation(filters)

        filter_list = []
        for filter_obj in filters:
            filter_list.append(cls.__recursive_map(filter_obj, operation))

        # Ensure the returned iterable is the same type as the original
        return type(filters)(filter_list)



"""
Decorator for making the authoring of "pure" components easier
"""

# Standard
from typing import Callable, Dict, Optional, Type

# Local
from .component import Component
from .controller import Controller


def component(name: str) -> Callable[[Type], Type]:
    """The @component decorator is the primary entrypoint for creating an
    oper8.Component. It ensures the wrapped type's interface matches the expected
    Component interface, including the "name" class attribute.

    Args:
        name:  str
            The name string will be set as the class property for the wrapped
            class

    Returns:
        decorator:  Callable[[Type[Component]], Type[Component]]
            The decorator function that will be invoked on construction of
            decorated classes
    """

    def decorator(cls: Type[Component]) -> Type[Component]:
        cls.name = name
        return cls

    return decorator


def controller(  # pylint: disable=too-many-arguments
    group: str,
    version: str,
    kind: str,
    finalizer: str = None,
    extra_properties: Optional[Dict[str, any]] = None,
) -> Callable[[Type[Controller]], Type[Controller]]:
    """The @controller decorator is the primary entrypoint for creating an
    oper8.Controller. It ensures the wrapped type's interface matches the
    required Controller interface, including class properties.

    NOTE: The `extra_properties` argument is an entrypoint for loosely coupled
        Controller-specific configuration that is tied to the specific
        WatchManager implementation being used. The current list of useful
        properties is:

        * disable_vcs: This can be used to tell the AnsibleWatchManager
            that the Controller will not use ansible-vcs, even if other
            Controllers managed by the same operator do.
        * pwm_filters: This can be used to tell the PythonWatchManager of any
            additional watch filters. If value is a list then the filters are added
            to all watches including dependent watches. If value is a dict than
            it expects the keys to be the resource global id with the values being a list
            of filters for that resource
        * pwm_subsystems: This can be used to tell the PythonWatchManager of any
            subsystem relations. This allows a "subsystem" controller to be ran during
            the reconciliation of another similar to the DryRunWatchManager

    Args:
        group:  str
            The apiVersion group for the resource this controller manages
        version:  str
            The apiVersion version for the resource this controller manages
        kind:  str
            The kind for the resource this controller manages
        extra_properties:  Optional[Dict[str, any]]
            Extra properties that should be defined as class-properties for this
            controller

    Returns:
        decorator:  Callable[[Type[Controller]], Type[Controller]]
            The decorator function that will be invoked on construction of
            decorated classes
    """

    def decorator(cls: Type[Controller]) -> Type[Controller]:
        cls.group = group
        cls.version = version
        cls.kind = kind
        for key, val in (extra_properties or {}).items():
            setattr(cls, key, val)
        if finalizer is not None:
            cls.finalizer = finalizer
        return cls

    return decorator



"""
Test the Node and ResourceNode classes functionality
"""

# Third Party
import pytest

# Local
from oper8.dag import Node, ResourceNode

################################################################################
## Node Tests #############################################################
################################################################################


def test_node_creation():
    """Test node initialization"""
    node = Node()
    assert node.get_name() == None
    assert node.get_data() == None
    node = Node("test", "data")
    assert node.get_name() == "test"
    assert node.get_data() == "data"


def test_node_children():
    """Test node children functions"""
    node_a = Node("a")
    node_b = Node("b")
    node_c = Node("c")
    node_a.add_child(node_b)
    node_a.add_child(node_c, "testdata")

    assert node_a.has_child(node_b)
    assert node_a.get_children() == [(node_b, None), (node_c, "testdata")]

    node_a.remove_child(node_b)
    assert not node_a.has_child(node_b)
    assert node_a.get_children() == [(node_c, "testdata")]


def test_node_cyclic_dependency():
    """Test node children functions"""
    node_a = Node("a")
    node_b = Node("b")
    node_a.add_child(node_b)
    with pytest.raises(ValueError):
        node_b.add_child(node_a)


def test_node_topology():
    """Test node topology function"""

    # Graph
    #  a->b->c
    node_a = Node("a")
    node_b = Node("b")
    node_c = Node("c")
    node_a.add_child(node_b)
    node_b.add_child(node_c)

    assert node_a.topology() == [node_c, node_b, node_a]
    assert node_b.topology() == [node_c, node_b]
    assert node_c.topology() == [node_c]


def test_node_dfs():
    """Test node dfs search"""

    # Graph
    #  a
    # / \
    # b  c
    # |
    # d
    node_a = Node("a")
    node_b = Node("b")
    node_c = Node("c")
    node_d = Node("d")
    node_a.add_child(node_b)
    node_a.add_child(node_c)
    node_b.add_child(node_d)

    # node a can reach all children
    assert node_a.dfs(node_b) and node_a.dfs(node_c) and node_a.dfs(node_d)
    # node b can only reach d and not c
    assert node_b.dfs(node_d) and not node_b.dfs(node_c)
    # node_d can't reach anything except itself
    assert (
        node_d.dfs(node_d)
        and not node_d.dfs(node_b)
        and not node_d.dfs(node_c)
        and not node_d.dfs(node_a)
    )


def test_node_equality():
    # Check equality
    node_a = Node("a")
    assert node_a != Node("b")
    assert node_a == Node("a")
    assert node_a != "arandomtype"

    # Check sorting
    assert node_a < Node("b")
    assert not Node("z") < node_a
    assert not Node("z") < "arandomtype"


def test_node_descriptors():
    assert hash(Node("a")) == hash("a")
    assert str(Node("a")) == "Node('a', None)"


################################################################################
## Resource Node Tests #############################################################
################################################################################


def create_dummy_kube_resource():
    return {
        "kind": "Foo",
        "apiVersion": "foo.bar/v1",
        "metadata": {
            "name": "foo",
            "namespace": "default",
        },
    }


def test_resource_node_attributes():
    node = ResourceNode("a", create_dummy_kube_resource())
    assert node.api_group == "foo.bar"
    assert node.api_version == "foo.bar/v1"
    assert node.kind == "Foo"
    assert node.metadata == {"name": "foo", "namespace": "default"}
    assert node.name == "foo"


def test_resource_node_add_dependency():
    node_a = ResourceNode("a", create_dummy_kube_resource())
    node_b = ResourceNode("b", create_dummy_kube_resource())
    node_a.add_dependency(node_b)
    assert node_a.has_child(node_b)
    assert node_a.get_children() == [(node_b, None)]



#!/usr/bin/env python
"""
The main module provides the executable entrypoint for oper8
"""

# Standard
from typing import Dict, Tuple
import argparse

# First Party
import aconfig
import alog

# Local
from .cmd import CheckHeartbeatCmd, CmdBase, RunOperatorCmd, SetupVCSCmd
from .component import config
from .config import library_config
from .log_format import Oper8JsonFormatter

## Constants ###################################################################

log = alog.use_channel("MAIN")

## Helpers #####################################################################


def add_library_config_args(parser, config_obj=None, path=None):
    """Automatically add args for all elements of the library config"""
    path = path or []
    setters = {}
    config_obj = config_obj or config
    for key, val in config_obj.items():
        sub_path = path + [key]

        # If this is a nested arg, recurse
        if isinstance(val, aconfig.AttributeAccessDict):
            sub_setters = add_library_config_args(parser, config_obj=val, path=sub_path)
            for dest_name, nested_path in sub_setters.items():
                setters[dest_name] = nested_path

        # Otherwise, add an argument explicitly
        else:
            arg_name = ".".join(sub_path)
            dest_name = "_".join(sub_path)
            kwargs = {
                "default": val,
                "dest": dest_name,
                "help": f"Library config override for {arg_name} (see oper8.config)",
            }
            if isinstance(val, list):
                kwargs["nargs"] = "*"
            elif isinstance(val, bool):
                kwargs["action"] = "store_true"
            else:
                type_name = None
                if val is not None:
                    type_name = type(val)
                kwargs["type"] = type_name

            if (
                f"--{arg_name}"
                not in parser._option_string_actions  # pylint: disable=protected-access
            ):
                parser.add_argument(f"--{arg_name}", **kwargs)
                setters[dest_name] = sub_path
    return setters


def update_library_config(args, setters):
    """Update the library config values based on the parsed arguments"""
    for dest_name, config_path in setters.items():
        config_obj = library_config
        while len(config_path) > 1:
            config_obj = config_obj[config_path[0]]
            config_path = config_path[1:]
        config_obj[config_path[0]] = getattr(args, dest_name)


def add_command(
    subparsers: argparse._SubParsersAction,
    cmd: CmdBase,
) -> Tuple[argparse.ArgumentParser, Dict[str, str]]:
    """Add the subparser and set up the default fun call"""
    parser = cmd.add_subparser(subparsers)
    parser.set_defaults(func=cmd.cmd)
    library_args = parser.add_argument_group("Library Configuration")
    library_config_setters = add_library_config_args(library_args)
    return parser, library_config_setters


## Main ########################################################################


def main():
    """The main module provides the executable entrypoint for oper8"""
    parser = argparse.ArgumentParser(description=__doc__)

    # Add the subcommands
    subparsers = parser.add_subparsers(help="Available commands", dest="command")
    run_operator_cmd = RunOperatorCmd()
    run_operator_parser, library_config_setters = add_command(
        subparsers, run_operator_cmd
    )
    run_health_check_cmd = CheckHeartbeatCmd()
    add_command(subparsers, run_health_check_cmd)
    setup_vcs_cmd = SetupVCSCmd()
    add_command(subparsers, setup_vcs_cmd)

    # Use a preliminary parser to check for the presence of a command and fall
    # back to the default command if not found
    check_parser = argparse.ArgumentParser(add_help=False)
    check_parser.add_argument("command", nargs="?")
    check_args, _ = check_parser.parse_known_args()
    if check_args.command not in subparsers.choices:
        args = run_operator_parser.parse_args()
    else:
        args = parser.parse_args()

    # Provide overrides to the library configs
    update_library_config(args, library_config_setters)

    # Reconfigure logging
    alog.configure(
        default_level=config.log_level,
        filters=config.log_filters,
        formatter=Oper8JsonFormatter() if config.log_json else "pretty",
        thread_id=config.log_thread_id,
    )

    # Run the command's function
    args.func(args)


if __name__ == "__main__":  # pragma: no cover
    main()



"""
Tests for functions in oper8.utils
"""

# Standard
import contextlib
import datetime
import io

# Third Party
import pytest

# Local
from oper8 import constants, utils
from oper8.dag import ResourceNode
from oper8.test_helpers.helpers import library_config, setup_cr, setup_session

## Helpers #####################################################################


def make_session_annotations(annotations):
    return setup_session(full_cr=setup_cr(metadata={"annotations": annotations}))


## get_passthrough_annotations #################################################


def test_get_passthrough_annotations_all_present():
    """Make sure that all expected annotations do get passed through"""
    annotations = {k: "foobar" for k in constants.PASSTHROUGH_ANNOTATIONS}
    passthroughs = utils.get_passthrough_annotations(
        make_session_annotations(annotations)
    )
    assert annotations == passthroughs

    # Since it's particularly important, explicitly make sure that the
    # config_defaults annotation is included
    assert constants.CONFIG_DEFAULTS_ANNOTATION_NAME in passthroughs


def test_get_passthrough_annotations_some_present():
    """Make sure that only the oper8 annotations present are passed through and
    new ones are not added
    """
    annotations = {
        constants.PAUSE_ANNOTATION_NAME: "False",
        constants.LOG_DEFAULT_LEVEL_NAME: "debug",
        constants.LOG_FILTERS_NAME: "FOO:off,BAR:debug4",
    }
    assert (
        utils.get_passthrough_annotations(make_session_annotations(annotations))
        == annotations
    )


def test_get_passthrough_annotations_ignore_others():
    """Make sure that non-oper8 annotations are not passed through"""
    annotations = {
        "other": "annotation",
        constants.PAUSE_ANNOTATION_NAME: "False",
    }
    passthroughs = utils.get_passthrough_annotations(
        make_session_annotations(annotations)
    )
    assert "other" not in passthroughs


def test_get_passthrough_annotations_no_temporary_patches():
    """Make sure that the TemporaryPatch annotation is NOT passed through"""
    annotations = {
        constants.TEMPORARY_PATCHES_ANNOTATION_NAME: "foobar",
    }
    passthroughs = utils.get_passthrough_annotations(
        make_session_annotations(annotations)
    )
    assert constants.TEMPORARY_PATCHES_ANNOTATION_NAME not in passthroughs


## nested_set / nested_get #####################################################


def test_nested_set_get():
    """Test nested_set and nested_get"""

    # Happy Path
    d = {}
    utils.nested_set(d, "foo.bar", 1)
    assert utils.nested_get(d, "foo.bar") == 1
    assert "foo" in d
    assert "bar" in d["foo"]
    assert d["foo"]["bar"] == 1

    # Bad intermediate key
    with pytest.raises(TypeError):
        utils.nested_set({"foo": 1}, "foo.bar", 1)
    with pytest.raises(TypeError):
        utils.nested_get({"foo": 1}, "foo.bar")

    # Get intermediate missing
    assert utils.nested_get({}, "foo.bar") is None
    assert utils.nested_get({}, "foo.bar", "default") == "default"
    assert utils.nested_get({"foo": {}}, "foo.bar", "default") == "default"


## classproperty decorators ####################################################


def test_classproperty():
    """Make sure that an attribute defined with @classproperty can be accessed
    without an instance of the class
    """

    class Foo:
        @utils.classproperty
        def bar(cls):
            return 1

    assert Foo.bar == 1


def test_abstractclassproperty():
    """Make sure that an attribute defined with @abstractclassproperty cannot be
    accessed without an implementation, but can be with a valid @classproperty
    implementation or a class member value
    """

    class FooBase:
        @utils.abstractclassproperty
        def bar(cls):
            """This will raise"""

        def __init__(self):
            self.bar

    class FooChildOne(FooBase):
        @utils.classproperty
        def bar(cls):
            return 1

    class FooChildTwo(FooBase):
        bar = 2

    class FooChildThree(FooBase):
        pass

    setattr(FooChildThree, "bar", 3)

    with pytest.raises(NotImplementedError):
        FooBase()

    assert FooChildOne.bar == 1
    assert FooChildTwo.bar == 2
    assert FooChildThree.bar == 3

    # Make sure help can be called on an abstractclassproperty without raising
    buf = io.StringIO()
    with contextlib.redirect_stdout(buf):
        help(FooBase.bar)
    buf.seek(0)
    help_str = buf.read()
    assert help_str


## sanitize_for_serialization #################################################


def get_generic_openapi_resource_type(provided_map=None):
    class OpenApiResource:
        attribute_map = provided_map

        @property
        def property() -> str:
            return "classproperty"

        def __init__(self, **kwargs):
            for k, v in kwargs.items():
                setattr(self, k, v)

    return OpenApiResource


def test_sanitize_for_serialization_types():
    """Make sure that the internal resource name is formatted correctly.
    This one is also pretty basic...
    """
    output_dict = (
        utils.sanitize_for_serialization(
            {
                "kind": "Foo",
                "metadata": {"name": "test"},
                "spec": {
                    "none": None,
                    "should_be_empty": {
                        "null": None,
                    },
                    "list": ["listitem"],
                    "tuple": ("tupleitem",),
                    "date": datetime.datetime(2020, 1, 1),
                    "resourceNode": ResourceNode(
                        name="test", manifest={"metadata": {"name": "test"}}
                    ),
                    "classProperty": get_generic_openapi_resource_type().property,
                    "openapiType": get_generic_openapi_resource_type(
                        {
                            "api_version": "apiVersion",
                            "kind": "kind",
                            "metadata": "metadata",
                            "spec": "spec",
                            "status": "status",
                        }
                    )(
                        api_version="v1",
                        kind="Test",
                        spec={"container": []},
                        metadata={"name": "test"},
                        status=get_generic_openapi_resource_type(
                            {"reconciled_version": "reconciledVersion"}
                        )(reconciled_version=1),
                    ),
                },
            },
        ),
    )
    assert output_dict[0] == {
        "kind": "Foo",
        "metadata": {"name": "test"},
        "spec": {
            "date": "2020-01-01T00:00:00",
            "list": ["listitem"],
            "classProperty": "classproperty",
            "openapiType": {
                "apiVersion": "v1",
                "kind": "Test",
                "metadata": {"name": "test"},
                "spec": {"container": []},
                "status": {"reconciledVersion": 1},
            },
            "resourceNode": {"metadata": {"name": "test"}},
            "should_be_empty": {},
            "tuple": ("tupleitem",),
        },
    }


## Manifest Version #################################################


def test_get_manifest_version():
    cr = setup_cr(version="1.2.3")
    assert utils.get_manifest_version(cr) == "1.2.3"

    with library_config(vcs={"version_override": "2.0.0"}):
        assert utils.get_manifest_version(cr) == "2.0.0"


## Finalizers #################################################


@pytest.mark.parametrize(
    ["cr", "finalizer", "expected_finalizers"],
    [
        [setup_cr(), "testfinalizer", ["testfinalizer"]],
        [
            setup_cr(metadata={"finalizers": ["otherfinalizer"]}),
            "testfinalizer",
            ["testfinalizer", "otherfinalizer"],
        ],
        [
            setup_cr(metadata={"finalizers": ["testfinalizer"]}),
            "testfinalizer",
            ["testfinalizer"],
        ],
    ],
)
def test_add_finalizer(cr, finalizer, expected_finalizers):
    session = setup_session(full_cr=cr)

    utils.add_finalizer(session, finalizer)

    assert set(session.finalizers) == set(expected_finalizers)

    success, obj = session.get_object_current_state(
        session.kind, session.name, session.api_version
    )
    assert success
    assert set(obj["metadata"]["finalizers"]) == set(expected_finalizers)


@pytest.mark.parametrize(
    ["cr", "finalizer", "expected_finalizers"],
    [
        [setup_cr(), "testfinalizer", []],
        [setup_cr(metadata={"finalizers": ["testfinalizer"]}), "testfinalizer", []],
        [
            setup_cr(metadata={"finalizers": ["testfinalizer", "otherfinalizer"]}),
            "testfinalizer",
            ["otherfinalizer"],
        ],
    ],
)
def test_remove_finalizer(cr, finalizer, expected_finalizers):
    session = setup_session(full_cr=cr)

    utils.remove_finalizer(session, finalizer)

    assert set(session.finalizers) == set(expected_finalizers)

    success, obj = session.get_object_current_state(
        session.kind, session.name, session.api_version
    )
    assert success
    assert set(obj["metadata"]["finalizers"]) == set(expected_finalizers)


def test_remove_finalizer_not_found():
    """Make sure that when the resource has already been removed from the
    cluster, there is no error on remove_finalizer
    """
    session = setup_session(full_cr=setup_cr(), deploy_initial_cr=False)
    utils.remove_finalizer(session, "testfinalizer")
    success, obj = session.get_object_current_state(
        session.kind, session.name, session.api_version
    )
    assert success
    assert not obj



"""
This module defines the interface needed to provide TLS key/cert pairs to a
given microservice and fetch a client-side certificate for making calls to a
microservice that serves a key/cert pair derived from this context.
"""

# Standard
from typing import List, Tuple
import abc

# First Party
import aconfig

# Local
from oper8 import Component, Session


class ITlsContext(abc.ABC):
    """This interface encapsulates the management of TLS for a running instance
    of the operand. It encapsulates the following functions:

    * Manage a CA key/cert pair for signing derived microservice certificates
    * Create derived key/cert pairs for individual microservices
    """

    # The string constant that will be used by derived classes to define the
    # type label string
    _TYPE_LABEL_ATTRIBUTE = "TYPE_LABEL"

    ## Construction ############################################################

    def __init__(self, session: Session, config: aconfig.Config):
        """Construct with the current session so that member functions do not
        take it as an argument

        Args:
            session:  Session
                The current deploy session
            config:  aconfig.Config
                The config for this instance
        """
        self._session = session
        self._config = config

    @property
    def session(self) -> Session:
        return self._session

    @property
    def config(self) -> aconfig.Config:
        return self._config

    ## Interface ###############################################################

    def request_server_key_cert_pair(  # noqa: B027
        self,
        server_component: Component,
        san_hostnames_list: List[str],
        san_ip_list: List[str] = None,
        key_name: str = None,
        intermediate_ca: bool = False,
    ) -> None:
        """Request creation of the PEM encoded value of the key/cert pair for a
        given server. This function has to be called from before render_chart is
        called. I.e., parse_config / Component constructor phase.
        Implementations of this function will generate the pair (in background)
        if it has not been already requested.

        Args:
            server_component:  Component
                The Component that manages the server. This can be used to add
                a new Component if needed that will manage the resource for the
                derived content and configure dependencies.
            san_hostnames_list:  List[str]
                The list of Subject Alternate Names (hostnames only)
            san_ip_list:  List[str]
                The list of Subject Alternate Names (ip addresses only, IPv4,
                IPv6)
            key_name:  str
                In case component requires multiple certificates. The key_name
                is used to distinguishes between component cert requests.
            intermediate_ca:  bool
                Whether or not to configure the certificate for use as an
                intermediate CA. This implies setting the key_cert_sign usage
                bit in the generated cert.
                Reference: https://ldapwiki.com/wiki/KeyUsage
        """

    @abc.abstractmethod
    def get_server_key_cert_pair(
        self,
        server_component: Component,
        key_name: str = None,
        encode: bool = True,
        existing_key_pem: str = None,
        existing_cert_pem: str = None,
    ) -> Tuple[str, str]:
        """Get the PEM encoded value of the key/cert pair for a given server.
        You have to forst request_server_key_cert_pair in render_config phase,
         and later in render_chart retrieve generated cert.

        Args:
            server_component:  Component
                The Component that manages the server. This can be used to add
                a new Component if needed that will manage the resource for the
                derived content and configure dependencies.
            key_name:  str
                In case component requires multiple certificates. The key_name
                is used to distinguies between component cert requests.
            encode:  bool
                Whether or not to base64 encode the output pem strings
            existing_key_pem: str
                Optionaly, you may provide the (decoded) value of PK/CERK pair.
                TLS context is free to check the Cert/PK and return this pair or
                generate new one.
            existing_cert_pem: str,
                Optionaly, you may provide the (decoded) value of PK/CERK pair.
                TLS context is free to check the Cert/PK and return this pair or
                generate new one.
        Returns:
            key_pem:  Optional[str]
                This is the pem-encoded key content (base64
                encoded if encode is set)
            cert_pem:  Optional[str]
                This is the pem-encoded cert content (base64
                encoded if encode is set)
        """

    @abc.abstractmethod
    def get_client_cert(
        self,
        client_component: Component,
        encode: bool = True,
    ) -> str:
        """Get a cert which can be used by a client to connect to a server which
        is serving using a key/cert pair signed by the shared CA.

        Args:
            client_component:  Component
                The Component that manages the client. This can be used to add
                a new Component if needed that will manage the resource for the
                derived content and configure dependencies.
            encode:  bool
                Whether or not to base64 encode the output pem strings

        Returns:
            crt_pem:  Optional[str]
                The pem-encoded cert (base64 encoded if encode set).
        """



"""
Tests for the ReconcileThread
"""
# Standard
from datetime import timedelta
import time

# Third Party
import pytest

# Local
from oper8.deploy_manager.dry_run_deploy_manager import DryRunDeployManager
from oper8.deploy_manager.kube_event import KubeEventType
from oper8.reconcile import ReconciliationResult, RequeueParams
from oper8.test_helpers.helpers import library_config
from oper8.test_helpers.pwm_helpers import MockedReconcileThread, make_managed_object
from oper8.watch_manager.python_watch_manager.utils.types import (
    ReconcileRequest,
    ReconcileRequestType,
    ResourceId,
    WatchRequest,
)

## Helpers #####################################################################


@pytest.mark.timeout(5)
def test_reconcile_thread_happy_path():
    dm = DryRunDeployManager()
    result = ReconciliationResult(requeue=False)
    reconcile_thread = MockedReconcileThread(
        deploy_manager=dm, returned_messages=[[result]]
    )

    resource = make_managed_object()
    dm.deploy([resource.definition])
    reconcile_thread.start_thread()

    request = ReconcileRequest(
        controller_type=None, type=KubeEventType.ADDED, resource=resource
    )
    reconcile_thread.push_request(request)

    time.sleep(3)
    reconcile_thread.stop_thread()
    # Make sure the process was started
    assert reconcile_thread.processes_started == 1
    assert reconcile_thread.processes_finished == 1
    # Make sure a periodic timer event was created
    assert reconcile_thread.timer_events.qsize() == 1
    assert (
        reconcile_thread.timer_events.get().args[0].type
        == ReconcileRequestType.PERIODIC
    )


@pytest.mark.timeout(5)
def test_reconcile_requeue():
    dm = DryRunDeployManager()
    requeue_result = ReconciliationResult(
        requeue=True, requeue_params=RequeueParams(requeue_after=timedelta(seconds=1))
    )
    end_result = ReconciliationResult(requeue=False)
    reconcile_thread = MockedReconcileThread(
        deploy_manager=dm,
        subprocess_wait_time=0,
        returned_messages=[[requeue_result], [end_result]],
    )

    resource = make_managed_object()
    dm.deploy([resource.definition])
    reconcile_thread.start_thread()

    request = ReconcileRequest(
        controller_type=None, type=KubeEventType.ADDED, resource=resource
    )
    reconcile_thread.push_request(request)

    time.sleep(3)
    reconcile_thread.stop_thread()
    # Make sure the process was started
    assert reconcile_thread.processes_started == 2
    assert reconcile_thread.processes_finished == 2
    # Make sure a periodic timer event was created
    assert reconcile_thread.timer_events.qsize() == 2
    assert (
        reconcile_thread.timer_events.get().args[0].type
        == ReconcileRequestType.REQUEUED
    )
    assert (
        reconcile_thread.timer_events.get().args[0].type
        == ReconcileRequestType.PERIODIC
    )


@pytest.mark.timeout(5)
def test_reconcile_overflow():
    dm = DryRunDeployManager()
    result = ReconciliationResult(requeue=False)
    with library_config(
        python_watch_manager={"process_context": "fork", "max_concurrent_reconciles": 2}
    ):
        reconcile_thread = MockedReconcileThread(
            deploy_manager=dm, subprocess_wait_time=2, returned_messages=[[result]]
        )

        resource = make_managed_object()
        dm.deploy([resource.definition])
        first_request = ReconcileRequest(
            controller_type=None, type=KubeEventType.ADDED, resource=resource
        )
        resource = make_managed_object(name="second")
        dm.deploy([resource.definition])
        second_request = ReconcileRequest(
            controller_type=None, type=KubeEventType.ADDED, resource=resource
        )
        resource = make_managed_object(name="overflow")
        dm.deploy([resource.definition])
        overflow_request = ReconcileRequest(
            controller_type=None, type=KubeEventType.ADDED, resource=resource
        )
        # Push requests before starting thread to fill queue before it can be
        # emptied
        reconcile_thread.push_request(first_request)
        reconcile_thread.push_request(second_request)
        reconcile_thread.push_request(overflow_request)
        reconcile_thread.start_thread()

        # Stop thread and assert only 2 processes were started
        time.sleep(0.5)
        reconcile_thread.stop_thread()
        assert reconcile_thread.processes_started == 2


@pytest.mark.timeout(8)
def test_reconcile_overflow_handled():
    dm = DryRunDeployManager()
    result = ReconciliationResult(requeue=False)
    with library_config(
        python_watch_manager={"process_context": "fork", "max_concurrent_reconciles": 1}
    ):
        reconcile_thread = MockedReconcileThread(
            deploy_manager=dm, subprocess_wait_time=0, returned_messages=[[result]]
        )

        resource = make_managed_object()
        dm.deploy([resource.definition])
        first_request = ReconcileRequest(
            controller_type=None, type=KubeEventType.ADDED, resource=resource
        )
        resource = make_managed_object(name="second")
        dm.deploy([resource.definition])
        second_request = ReconcileRequest(
            controller_type=None, type=KubeEventType.ADDED, resource=resource
        )
        resource = make_managed_object(name="overflow")
        dm.deploy([resource.definition])
        third_request = ReconcileRequest(
            controller_type=None, type=KubeEventType.ADDED, resource=resource
        )
        # Push requests before starting thread to fill queue before it can be
        # emptied
        reconcile_thread.push_request(first_request)
        reconcile_thread.push_request(second_request)
        reconcile_thread.push_request(third_request)
        reconcile_thread.start_thread()

        # Stop thread and assert that the reconcile thread was able to handle all requests
        time.sleep(5)
        reconcile_thread.stop_thread()
        assert reconcile_thread.processes_started == 3


@pytest.mark.timeout(5)
def test_reconcile_pending():
    dm = DryRunDeployManager()
    result = ReconciliationResult(requeue=False)
    reconcile_thread = MockedReconcileThread(
        deploy_manager=dm,
        subprocess_wait_time=1,
        returned_messages=[[result], [result], [result]],
    )

    resource = make_managed_object()
    dm.deploy([resource.definition])
    add_request = ReconcileRequest(
        controller_type=None, type=KubeEventType.ADDED, resource=resource
    )
    modified_request = ReconcileRequest(
        controller_type=None, type=KubeEventType.MODIFIED, resource=resource
    )
    duplicate_modified_request = ReconcileRequest(
        controller_type=None, type=KubeEventType.MODIFIED, resource=resource
    )
    deleted_requested = ReconcileRequest(
        controller_type=None, type=KubeEventType.DELETED, resource=resource
    )
    # Push requests before starting thread to fill queue
    reconcile_thread.push_request(add_request)
    reconcile_thread.push_request(modified_request)
    reconcile_thread.push_request(duplicate_modified_request)
    reconcile_thread.push_request(deleted_requested)
    reconcile_thread.start_thread()

    # Stop thread and assert that the reconcile thread combined all requests
    time.sleep(3)
    reconcile_thread.stop_thread()
    assert reconcile_thread.processes_started == 2
    assert reconcile_thread.processes_finished == 2


@pytest.mark.timeout(5)
def test_reconcile_watch_request():
    dm = DryRunDeployManager()

    result = ReconciliationResult(requeue=False)
    resource = make_managed_object()
    dm.deploy([resource.definition])

    watch_request = WatchRequest(
        watched=ResourceId.from_resource(resource),
        requester=ResourceId.from_resource(resource),
        controller_type=None,
    )
    reconcile_thread = MockedReconcileThread(
        deploy_manager=dm,
        subprocess_wait_time=0,
        returned_messages=[[result, watch_request]],
    )
    reconcile_thread.start_thread()

    request = ReconcileRequest(
        controller_type=None, type=KubeEventType.ADDED, resource=resource
    )
    # Push requests before starting thread to fill queue
    reconcile_thread.push_request(request)

    # Stop thread and assert that the reconcile thread combined all requests
    time.sleep(2)
    reconcile_thread.stop_thread()
    assert reconcile_thread.processes_started == 1
    assert reconcile_thread.processes_finished == 1
    assert reconcile_thread.watch_threads_created == 1



"""
Dry run implementation of the WatchManager abstraction
"""

# Standard
from functools import partial
from typing import Optional, Type
import logging

# First Party
import alog

# Local
from ..controller import Controller
from ..deploy_manager import DryRunDeployManager
from ..reconcile import ReconcileManager
from .base import WatchManagerBase

log = alog.use_channel("DRWAT")


class DryRunWatchManager(WatchManagerBase):
    """
    The DryRunWatchManager implements the WatchManagerBase interface with using
    a single shared DryRunDeployManager to manage an in-memory representation of
    the cluster.
    """

    reconcile_manager = None

    def __init__(
        self,
        controller_type: Type[Controller],
        deploy_manager: Optional[DryRunDeployManager] = None,
    ):
        """Construct with the type of controller to watch and optionally a
        deploy_manager instance. A deploy_manager will be constructed if none is
        given.

        Args:
            controller_type:  Type[Controller]
                The class for the controller that will be watched
            deploy_manager:  Optional[DryRunWatchManager]
                If given, this deploy_manager will be used. This allows for
                there to be pre-populated resources. Note that it _must_ be a
                DryRunDeployManager (or child class) that supports registering
                watches.
        """
        super().__init__(controller_type)

        # Set up the deploy manager
        self._deploy_manager = deploy_manager or DryRunDeployManager()

        # We lazily initialize the controller instance in watch and _resource in run_reconcile
        self._controller = None
        self._resource = {}

        # We initialize the reconcile_manager instance on first watch creation
        if not self.reconcile_manager:
            self.reconcile_manager = ReconcileManager(
                deploy_manager=self._deploy_manager, reimport_controller=False
            )

    def watch(self) -> bool:
        """Register the watch with the deploy manager"""
        if self._controller is not None:
            log.warning("Cannot watch multiple times!")
            return False

        log.debug("Registering %s with the DeployManager", self.controller_type)

        # Construct controller
        self._controller = self.controller_type()

        # Register watch and finalizers
        api_version = f"{self.group}/{self.version}"
        self._deploy_manager.register_watch(
            api_version=api_version,
            kind=self.kind,
            callback=partial(self.run_reconcile, False),
        )
        if self.controller_type.has_finalizer:
            log.debug("Registering finalizer")
            self._deploy_manager.register_finalizer(
                api_version=api_version,
                kind=self.kind,
                callback=partial(self.run_reconcile, True),
            )

        return True

    def wait(self):
        """There is nothing to do in wait"""

    def stop(self):
        """There is nothing to do in stop"""

    def run_reconcile(self, is_finalizer: bool, resource: dict):
        """Wrapper function to simplify parameter/partial mapping"""
        if not self.reconcile_manager:
            return

        # Only run reconcile if it's a unique resource
        resource_metadata = self._resource.get("metadata", {})
        if (
            self._resource.get("kind") == resource.get("kind")
            and self._resource.get("apiVersion") == resource.get("apiVersion")
            and resource_metadata.get("name")
            == resource.get("metadata", {}).get("name")
            and resource_metadata.get("namespace")
            == resource.get("metadata", {}).get("namespace")
        ):
            return

        # Save the current resource and log handlers then restore it after the reconcile
        # is completed
        log_formatters = {}
        for handler in logging.getLogger().handlers:
            log_formatters[handler] = handler.formatter
        current_resource = self._resource
        self._resource = resource

        self.reconcile_manager.reconcile(self._controller, resource, is_finalizer)
        self._resource = current_resource
        for handler, formatter in log_formatters.items():
            handler.setFormatter(formatter)



"""Common functions used for interacting with filters including default filter classes"""

# Standard
from functools import lru_cache
from typing import List, Type
import importlib
import inspect

# First Party
import alog

# Local
from .... import config
from ....exceptions import ConfigError
from .filters import (
    AnnotationFilter,
    CreationDeletionFilter,
    EnableFilter,
    Filter,
    GenerationFilter,
    NoGenerationFilter,
    PauseFilter,
    ResourceVersionFilter,
    UserAnnotationFilter,
)
from .manager import AndFilter, OrFilter

log = alog.use_channel("PWMFLTCOM")


### Factory Filter Classes

# Usable Default Filter Classes.
DEFAULT_FILTER_CLASS = AndFilter(
    CreationDeletionFilter,
    GenerationFilter,
    NoGenerationFilter,
    ResourceVersionFilter,
    PauseFilter,
)
ANNOTATION_FILTER_CLASS = OrFilter(DEFAULT_FILTER_CLASS, AnnotationFilter)
USER_ANNOTATION_FILTER_CLASS = OrFilter(DEFAULT_FILTER_CLASS, UserAnnotationFilter)

FILTER_CLASSES = {
    "default": DEFAULT_FILTER_CLASS,
    "annotation": ANNOTATION_FILTER_CLASS,
    "user-annotation": USER_ANNOTATION_FILTER_CLASS,
}


# Forward Declarations
CONTROLLER_TYPE = "Controller"
CONTROLLER_CLASS_TYPE = Type[CONTROLLER_TYPE]
RESOURCE_ID_TYPE = "ResourceId"

### Factory Filter Functions


# Only compute the filters once to avoid reimporting/regathering
@lru_cache(maxsize=1)
def get_configured_filter() -> Filter:
    """Get the default filter that should be applied to every resource

    Returns:
        default_filter: Filter
            The default filter specified in the Config"""

    filter_name = config.python_watch_manager.filter

    # Check for filter in default list or attempt to
    # manually import one
    if filter_name in FILTER_CLASSES:
        filter_obj = FILTER_CLASSES[filter_name]
    elif inspect.isclass(filter_name) and issubclass(filter_name, Filter):
        filter_obj = filter_name
    elif isinstance(filter_name, str):
        filter_obj = import_filter(filter_name)
    # If no filter is provided then always enable
    else:
        filter_obj = EnableFilter

    log.debug2(f"Found filter: {filter_obj}")
    return filter_obj


def get_filters_for_resource_id(
    controller_type: CONTROLLER_CLASS_TYPE, resource_id: RESOURCE_ID_TYPE
) -> List[Filter]:
    """Get the filters for a particular resource_id given a controller_type

    Args:
        controller_type: CONTROLLER_CLASS_TYPE
            The controller type whose filters we're inspecting
        resource_id: "ResourceId"
            The requested resource

    Returns:
        filter_list: List[Filter]
            The list of filters to be applied
    """
    filters = getattr(controller_type, "pwm_filters", [])

    if isinstance(filters, list):
        return_filters = filters

    elif isinstance(filters, dict):
        return_filters = filters.get(resource_id.global_id, [])

    else:
        raise ConfigError(f"Invalid type for PWM filters: {type(filters)}")

    log.debug3(f"Found filters {return_filters} for resource: {resource_id}")
    return return_filters


### Helper Functions


def import_filter(filter_name: str) -> Filter:
    """Import a filter from a string reference

    Args:
        filter_name: str
            Filter name in <module>.<filter> form

    Returns:
        imported_filter: Filter
            The filter that was requested
    """
    module_path, class_name = filter_name.rsplit(".", 1)
    try:
        filter_module = importlib.import_module(module_path)
        filter_obj = getattr(filter_module, class_name)
    except (ImportError, AttributeError) as exc:
        raise ConfigError(
            f"Invalid Filter: {filter_name}. Module or class not found"
        ) from exc

    if (
        inspect.isclass(filter_obj) and not issubclass(filter_obj, Filter)
    ) and not isinstance(filter_obj, (Filter, list, tuple)):
        raise ConfigError(f"{filter_obj} is not a instance of {Filter}")

    return filter_obj



"""
COS instance factory
"""

# Local
from ..factory_base import DatastoreSingletonFactoryBase
from .connection import CosConnection


class CosFactory(DatastoreSingletonFactoryBase):
    """The common factory that will manage instances of COS for each deploy"""

    DATASTORE_TYPE = "cos"
    CONNECTION_TYPE = CosConnection



"""
Shared utilities for the PythonWatchManager
"""
# Standard
from datetime import timedelta
from typing import Any, List, Optional
import json
import logging
import pathlib
import platform
import re

# First Party
import alog

# Local
from .... import config

log = alog.use_channel("PWMCMMN")


## Time Functions

# Shamelessly stolen from
# https://stackoverflow.com/questions/4628122/how-to-construct-a-timedelta-object-from-a-simple-string
regex = re.compile(
    r"^((?P<hours>\d+?)hr)?((?P<minutes>\d+?)m)?((?P<seconds>\d*\.?\d+?)s)?$"
)


def parse_time_delta(
    time_str: str,
) -> Optional[timedelta]:  # pylint: disable=inconsistent-return-statements
    """Parse a string into a timedelta. Excepts values in the
    following formats: 1h, 5m, 10s, etc

    Args:
        time_str: str
            The string representation of a timedelta

    Returns:
        result: Optional[timedelta]
            The parsed timedelta if one could be found
    """
    parts = regex.match(time_str)
    if not parts or all(part is None for part in parts.groupdict().values()):
        return None
    parts = parts.groupdict()
    time_params = {}
    for name, param in parts.items():
        if param:
            time_params[name] = float(param)
    return timedelta(**time_params)


## Identity Util Functions


def get_operator_namespace() -> str:
    """Get the current namespace from a kubernetes file or config"""
    # Default to in cluster namespace file
    namespace_file = pathlib.Path(
        "/var/run/secrets/kubernetes.io/serviceaccount/namespace"
    )
    if namespace_file.is_file():
        return namespace_file.read_text(encoding="utf-8")
    return config.python_watch_manager.lock.namespace


def get_pod_name() -> str:
    """Get the current pod from env variables, config, or hostname"""

    pod_name = config.pod_name
    if not pod_name:
        log.warning("Pod name not detected, falling back to hostname")
        pod_name = platform.node().split(".")[0]

    return pod_name


## Helper functions


def obj_to_hash(obj: Any) -> str:
    """Get the hash of any jsonable python object

    Args:
        obj: Any
            The object to hash

    Returns:
        hash: str
            The hash of obj
    """
    return hash(json.dumps(obj, sort_keys=True))


def get_logging_handlers() -> List[logging.Handler]:
    """Get the current logging handlers"""
    logger = logging.getLogger()
    if not logger.handlers:
        handler = logging.StreamHandler()
        logger.addHandler(handler)

    return logger.handlers



"""
Python-based implementation of the WatchManager
"""

# Standard
from typing import List, Optional, Type
import threading

# First Party
import alog

# Local
from ... import config
from ...controller import Controller
from ...deploy_manager import OpenshiftDeployManager
from ..base import WatchManagerBase
from .filters import get_filters_for_resource_id
from .leader_election import LeadershipManagerBase, get_leader_election_class
from .threads import HeartbeatThread, ReconcileThread, WatchThread
from .threads.watch import create_resource_watch, get_resource_watches
from .utils import ResourceId, WatchRequest

log = alog.use_channel("PYTHW")


class PythonWatchManager(WatchManagerBase):
    """The PythonWatchManager uses the kubernetes watch client to watch
    a particular Controller and execute reconciles. It does the following
    two things

    1. Request a generic watch request for each namespace
    2. Start a reconcile thread to start reconciliation subprocesses
    """

    def __init__(
        self,
        controller_type: Type[Controller],
        deploy_manager: Optional[OpenshiftDeployManager] = None,
        namespace_list: Optional[List[str]] = None,
    ):
        """Initialize the required threads and submit the watch requests
        Args:
            controller_type: Type[Controller]
                The controller to be watched
            deploy_manager: Optional[OpenshiftDeployManager] = None
                An optional DeployManager override
            namespace_list: Optional[List[str]] = []
                A list of namespaces to watch
        """
        super().__init__(controller_type)

        # Handle functional args
        if deploy_manager is None:
            log.debug("Using OpenshiftDeployManager")
            deploy_manager = OpenshiftDeployManager()
        self.deploy_manager = deploy_manager

        # Setup watch namespace
        self.namespace_list = namespace_list or []
        if not namespace_list and config.watch_namespace != "":
            self.namespace_list = config.watch_namespace.split(",")

        # Setup Control variables
        self.shutdown = threading.Event()

        # Setup Threads. These are both singleton instances and will be
        # the same across all PythonWatchManagers
        self.leadership_manager: LeadershipManagerBase = get_leader_election_class()(
            self.deploy_manager
        )
        self.reconcile_thread: ReconcileThread = ReconcileThread(
            deploy_manager=self.deploy_manager,
            leadership_manager=self.leadership_manager,
        )
        self.heartbeat_thread: Optional[HeartbeatThread] = None
        if config.python_watch_manager.heartbeat_file:
            self.heartbeat_thread = HeartbeatThread(
                config.python_watch_manager.heartbeat_file,
                config.python_watch_manager.heartbeat_period,
            )

        # Start thread for each resource watch
        self.controller_watches: List[WatchThread] = []
        if len(self.namespace_list) == 0 or "*" in self.namespace_list:
            self.controller_watches.append(self._add_resource_watch())
        else:
            for namespace in self.namespace_list:
                self.controller_watches.append(self._add_resource_watch(namespace))

    ## Interface ###############################################################

    def watch(self) -> bool:
        """Check for leadership and start all threads

        Returns:
            success:  bool
                True if all threads process are running correctly
        """
        log.info("Starting PythonWatchManager: %s", self)

        if not self.leadership_manager.is_leader():
            log.debug("Acquiring Leadership lock before starting %s", self)
            self.leadership_manager.acquire()

        # If watch has been shutdown then exit before starting threads
        if self.shutdown.is_set():
            return False

        # Start reconcile thread and all watch threads
        self.reconcile_thread.start_thread()
        for watch_thread in self.controller_watches:
            log.debug("Starting watch_thread: %s", watch_thread)
            watch_thread.start_thread()
        if self.heartbeat_thread:
            log.debug("Starting heartbeat_thread")
            self.heartbeat_thread.start_thread()
        return True

    def wait(self):
        """Wait shutdown to be signaled"""
        self.shutdown.wait()

    def stop(self):
        """Stop all threads. This waits for all reconciles
        to finish
        """

        log.info(
            "Stopping PythonWatchManager for %s/%s/%s",
            self.group,
            self.version,
            self.kind,
        )

        # Set shutdown and acquire leadership to clear any deadlocks
        self.shutdown.set()
        self.leadership_manager.acquire(force=True)

        # Stop all threads
        for watch in get_resource_watches():
            watch.stop_thread()
        self.reconcile_thread.stop_thread()
        self.leadership_manager.release()
        if self.heartbeat_thread:
            self.heartbeat_thread.stop_thread()

    ## Helper Functions ###############################################################

    def _add_resource_watch(self, namespace: Optional[str] = None):
        """Request a generic watch request. Optionally for a specific namespace

        Args:
            namespace: Optional[str] = None
                An optional namespace to watch
        """
        log.debug3("Adding %s request for %s", namespace if namespace else "", self)

        # In the global watch manager the controller is both
        # the watched and the requested objects
        resource_id = ResourceId.from_controller(self.controller_type, namespace)
        request = WatchRequest(
            controller_type=self.controller_type,
            watched=resource_id,
            requester=resource_id,
            filters=get_filters_for_resource_id(self.controller_type, resource_id),
        )
        return create_resource_watch(
            request,
            self.reconcile_thread,
            self.deploy_manager,
            self.leadership_manager,
        )



"""
Tests for the replace_utils functionality
"""

# Standard

# Third Party
import pytest

# First Party
import alog

# Local
from oper8.deploy_manager.replace_utils import requires_replace, REPLACE_FUNCS

## Helpers #####################################################################

log = alog.use_channel("TEST")


def sample_object():
    return {
        "original_value": "original",
        "envs": [
            {
                "name": "first",
                "value": "True",
            },
            {
                "name": "second",
                "valueFrom": "False",
            },
        ],
        "list": [
            {"name": "container1"},
            {"name": "container2"},
        ],
    }


## Replace functions ##################################################################


@pytest.mark.parametrize(
    ["desired_obj"],
    [
        [
            {
                "envs": [
                    {
                        "name": "first",
                        "valueFrom": "True",
                    },
                    {
                        "name": "second",
                        "valueFrom": "False",
                    },
                ],
            }
        ],
        [
            {
                "envs": [
                    {
                        "name": "first",
                        "value": "True",
                    },
                    {
                        "name": "second",
                        "value": "True",
                    },
                ],
            }
        ],
    ],
)
def test_value_operations(desired_obj):
    """Test that adding a ref to an object with none present adds as expected"""
    current_obj = sample_object()
    assert requires_replace(current_obj, desired_obj)
    # Ensure each replace function is still able to be called
    for func in REPLACE_FUNCS:
        func(current_obj, desired_obj)


@pytest.mark.parametrize(
    ["desired_obj"],
    [
        [
            {
                "list": [
                    {"name": "container1"},
                    {"name": "container2"},
                    {"name": "container3"},
                ],
            }
        ],
        [
            {
                "list": [
                    {"name": "container1"},
                ],
            }
        ],
        [
            {
                "list": [
                    {"name": "container1"},
                    {"name": "container_changed"},
                ],
            }
        ],
    ],
)
def test_list_operations(desired_obj):
    """Test that adding a ref to an object with none present adds as expected"""
    current_obj = sample_object()
    assert requires_replace(current_obj, desired_obj)
    # Ensure each replace function is still able to be called
    for func in REPLACE_FUNCS:
        func(current_obj, desired_obj)

## Patch functions ##################################################################


@pytest.mark.parametrize(
    ["desired_obj"],
    [
        [{"new_value": "patched"}],
        [{"added_list": [{"new_value"}]}],
        [{"original_value": "patched"}],
    ],
)
def test_patch_operations(desired_obj):
    current_obj = sample_object()
    assert not requires_replace(current_obj, desired_obj)



"""
The oper8.x module holds common implementations of reusable patterns built on
top of the abstractions in oper8. These are intended as reusable components that
can be share across many operator implementations.

One of the core principles of oper8 is that the schema for config is entirely up
to the user (with the _only_ exception being spec.version). In oper8.x, this is
not the case and there are many config conventions (CRD schema and backend) that
are encoded into the various utilities.
"""



"""
Shared utilities for managing TLS keys and certs
"""

# Standard
import base64
import datetime

# Third Party
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.serialization import Encoding, PrivateFormat
from cryptography.x509.oid import ExtendedKeyUsageOID, NameOID

# First Party
import alog

log = alog.use_channel("TLS")

DEFAULT_COMMON_NAME = "oper8.org"


def get_subject(common_name: str = DEFAULT_COMMON_NAME) -> x509.Name:
    """Get the subject object used when creating self-signed certificates. This
    will be consistent across all components, but will be tailored to the domain
    of the cluster.

    Args:
        common_name:  str
            The Common Name to use for this subject

    Returns:
        subject:  x509.Name
            The full subect object to use when constructing certificates
    """
    return x509.Name(
        [
            x509.NameAttribute(NameOID.COMMON_NAME, common_name),
        ]
    )


def generate_key(encode=True):
    """Generate a new RSA key for use when generating TLS components

    Args:
        encode:  bool
            Base64 encode the output pem bytes

    Returns:
        key:  RSAPrivateKey
            The key object that can be used to sign certificates
        key_pem:  str
            The PEM encoded string for the key
    """
    key = rsa.generate_private_key(
        public_exponent=65537, key_size=2048, backend=default_backend()
    )
    key_pem = key.private_bytes(
        Encoding.PEM,
        PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption(),
    )
    return (key, (base64.b64encode(key_pem) if encode else key_pem).decode("utf-8"))


def generate_ca_cert(key, encode=True):
    """Generate a Certificate Authority certificate based on a private key

    Args:
        key:  RSAPrivateKey
            The private key that will pair with this CA cert
        encode:  bool
            Base64 encode the output pem bytes

    Returns:
        ca:  str
            The PEM encoded string for this CA cert
    """

    # Create self-signed CA
    # The specifics of the extensions that are required for the CA were gleaned
    # from the etcd operator example found here:
    # https://github.com/openshift/etcd-ha-operator/blob/master/roles/tls_certs/templates/ca_crt_conf.j2
    log.debug("Creating CA")
    subject = get_subject()
    ca = (
        x509.CertificateBuilder()
        .subject_name(subject)
        .issuer_name(subject)
        .public_key(key.public_key())
        .serial_number(x509.random_serial_number())
        .not_valid_before(datetime.datetime.utcnow())
        .not_valid_after(
            # Our certificate will be valid for 10000 days
            datetime.datetime.utcnow()
            + datetime.timedelta(days=10000)
        )
        .add_extension(
            # X509v3 Basic Constraints: critical
            #     CA:TRUE
            x509.BasicConstraints(ca=True, path_length=None),
            critical=True,
        )
        .add_extension(
            # X509v3 Key Usage: critical
            #     Digital Signature, Key Encipherment, Certificate Sign
            x509.KeyUsage(
                digital_signature=True,
                content_commitment=False,
                key_encipherment=True,
                data_encipherment=False,
                key_agreement=False,
                key_cert_sign=True,
                crl_sign=False,
                encipher_only=False,
                decipher_only=False,
            ),
            critical=True,
        )
        .sign(key, hashes.SHA256(), default_backend())
    )

    cert_pem = ca.public_bytes(Encoding.PEM)
    return (base64.b64encode(cert_pem) if encode else cert_pem).decode("utf-8")


def generate_derived_key_cert_pair(ca_key, san_list, encode=True, key_cert_sign=False):
    """Generate a certificate for use in encrypting TLS traffic, derived from
    a common key

    Args:
        key:  RSAPrivateKey
            The private key that will pair with this CA cert
        san_list:  list(str)
            List of strings to use for the Subject Alternate Name
        encode:  bool
            Whether or not to base64 encode the output pem strings
        key_cert_sign:  bool
            Whether or not to set the key_cert_sign usage bit in the generated certificate.
            This may be needed when the derived key/cert will be used as an intermediate CA
            or expected to act as a self-signed CA.
            Reference: https://ldapwiki.com/wiki/KeyUsage

    Returns:
        key_pem:  str
            The pem-encoded key (base64 encoded if encode set)
        crt_pem:  str
            The pem-encoded cert (base64 encoded if encode set)
    """

    # Create a new private key for the server
    key, key_pem = generate_key(encode=encode)

    # Create the server certificate as if using a CSR. The final key will be
    # signed by the CA private key, but will have the public key from the
    # server's key.
    #
    # NOTE: It is not legal to use an identical Common Name for both the CA and
    #   the derived certificate. With openssl 1.1.1k, this results in an invalid
    #   certificate that fails with "self signed certificate."
    #   CITE: https://stackoverflow.com/a/19738223
    cert = (
        x509.CertificateBuilder()
        .subject_name(get_subject(f"{DEFAULT_COMMON_NAME}.server"))
        .issuer_name(get_subject())
        .public_key(key.public_key())
        .serial_number(x509.random_serial_number())
        .not_valid_before(datetime.datetime.utcnow())
        .not_valid_after(
            # Our certificate will be valid for 10000 days
            datetime.datetime.utcnow()
            + datetime.timedelta(days=10000)
        )
        .add_extension(
            x509.SubjectAlternativeName([x509.DNSName(san) for san in san_list]),
            critical=False,
        )
        .add_extension(
            # X509v3 Key Usage: critical
            #     Digital Signature, Key Encipherment
            x509.KeyUsage(
                digital_signature=True,
                content_commitment=False,
                key_encipherment=True,
                data_encipherment=False,
                key_agreement=False,
                key_cert_sign=key_cert_sign,
                crl_sign=False,
                encipher_only=False,
                decipher_only=False,
            ),
            critical=True,
        )
        .add_extension(
            # X509v3 Extended Key Usage:
            #     TLS Web Client Authentication, TLS Web Server Authentication
            x509.ExtendedKeyUsage(
                [ExtendedKeyUsageOID.CLIENT_AUTH, ExtendedKeyUsageOID.SERVER_AUTH]
            ),
            critical=False,
        )
        .sign(ca_key, hashes.SHA256(), default_backend())
    )

    crt_pem = cert.public_bytes(Encoding.PEM)
    return (key_pem, (base64.b64encode(crt_pem) if encode else crt_pem).decode("utf-8"))


def parse_private_key_pem(key_pem):
    """Parse the content of a pem-encoded private key file into an RSAPrivateKey

    Args:
        key_pem:  str
            The pem-encoded key (not base64 encoded)

    Returns:
        key:  RSAPrivateKey
            The parsed key object which can be used for signing certs
    """
    return serialization.load_pem_private_key(key_pem.encode("utf-8"), None)


def parse_public_key_pem_from_cert(cert_pem):
    """Extract the pem-encoded public key from a pem-encoded

    Args:
        cert_pem:  str
            The pem-encoded certificate (not base64 encoded)

    Returns:
        key:  RSAPrivateKey
            The parsed key object which can be used for signing certs
    """
    return (
        x509.load_pem_x509_certificate(cert_pem.encode("utf-8"))
        .public_key()
        .public_bytes(
            serialization.Encoding.PEM, serialization.PublicFormat.SubjectPublicKeyInfo
        )
        .decode("utf-8")
    )



"""
This implementation of the ITlsContext uses internal code to manage the tls
context
"""

# Standard
from typing import List, Optional, Tuple

# First Party
import alog

# Local
from ...oper8x_component import Oper8xComponent
from .. import common, tls
from .interface import ITlsContext
from oper8 import Component, Session, assert_cluster

log = alog.use_channel("TLSINT")


## InternalTlsContext ##########################################################


class InternalTlsContext(ITlsContext):
    __doc__ = __doc__

    TYPE_LABEL = "internal"

    def __init__(self, session: Session, *args, **kwargs):
        """At construct time, this instance will add a Component to the session
        which will manage the CA secret
        """
        super().__init__(session, *args, **kwargs)

        # Add the CA Component if it is not already present in the session.
        # There is a pretty nasty condition here when running in standalone mode
        # where the factory may attempt to recreate the singleton instance after
        # a subsystem has overwritten it, so we need to check and see if there
        # is a matching component in the session already
        pre_existing_component = [
            comp
            for comp in session.get_components()
            if comp.name == InternalCaComponent.name
        ]
        if pre_existing_component:
            self._component = pre_existing_component[0]
        else:
            self._component = InternalCaComponent(
                session=session, labels=self.config.labels
            )

        # Keep track of pairs for each server so that they are only generated
        # once
        self._server_pairs = {}

    ## Interface ###############################################################
    def request_server_key_cert_pair(
        self,
        server_component: Component,
        san_hostnames_list: List[str],
        san_ip_list: List[str] = None,
        key_name: str = None,
        intermediate_ca: bool = False,
    ) -> None:
        """Request creation of the PEM encoded value of the key/cert pair for a
        given server. This function has to be called from before render_chart is
        called. I.e., parse_config / Component constructor phase.
        Implementations of this function will generate the pair (in background)
        if it has not been already requested.

        Args:
            server_component:  Component
                The Component that manages the server. This can be used to add
                a new Component if needed that will manage the resource for the
                derived content and configure dependencies.
            san_list:  List[str]
                The list of Subject Alternate Names
            key_name:  str
                In case component requires multiple certificates. The key_name
                is used to distinguish between component cert requests.
            intermediate_ca:  bool
                Whether or not to configure the certificate for use as an
                intermediate CA. This implies setting the key_cert_sign usage
                bit in the generated cert.
                Reference: https://ldapwiki.com/wiki/KeyUsage
        """
        cache_key = server_component.name + (
            "-" + key_name if key_name is not None else ""
        )

        if cache_key in self._server_pairs:
            log.warning(
                "Certificate server key/cert pair for %s has been already "
                "requested. Ignoring this request.",
                cache_key,
            )
            return

        log.debug("Generating server key/cert pair for %s", cache_key)

        # Mark the server component as dependent on the internal component. This
        # is not strictly necessary since values are consumed by value, but it
        # makes sequential sense.
        self.session.add_component_dependency(server_component, self._component)

        # Get the CA's private key
        ca_key = tls.parse_private_key_pem(self._component.get_ca_key_cert()[0])

        # Generate the derived pair
        san_list = (san_hostnames_list or []) + (san_ip_list or [])
        self._server_pairs[cache_key] = tls.generate_derived_key_cert_pair(
            ca_key=ca_key,
            san_list=san_list,
            encode=False,
            key_cert_sign=intermediate_ca,
        )

    def get_server_key_cert_pair(
        self,
        server_component: Component,
        key_name: str = None,
        encode: bool = True,
        existing_key_pem: str = None,
        existing_cert_pem: str = None,
    ) -> Tuple[str, str]:
        """This function derives a server key/cert pair from the CA key/cert
        managed by the internal component.

        Args:
            server_component:  Component
                The Component that manages the server. This can be used to add
                a new Component if needed that will manage the resource for the
                derived content and configure dependencies.
            key_name:  str
                In case component requires multiple certificates. The key_name
                is used to distinguish between component cert requests.
            encode:  bool
                Whether or not to base64 encode the output pem strings
            existing_key_pem: str
                If both existing key/cert are specified, then they are returned
                immediately without any checks
            existing_cert_pem: str
                If both existing key/cert are specified, then they are returned
                immediately without any checks
         Returns:
            key_pem:  str
                This is the pem-encoded key content (base64
                encoded if encode is set)
            cert_pem:  str
                This is the pem-encoded cert content (base64
                encoded if encode is set)
        """
        log.debug2("Getting server key/cert pair for %s", server_component)
        cache_key = server_component.name + (
            "-" + key_name if key_name is not None else ""
        )

        assert (
            cache_key in self._server_pairs
        ), f"Trying to obtain certificate {key_name} which was not previouly requested"
        if existing_key_pem is not None and existing_cert_pem is not None:
            key_pem = existing_key_pem
            cert_pem = existing_cert_pem
        else:
            # Return the stored pair for this server
            (key_pem, cert_pem) = self._server_pairs[cache_key]
        if encode:
            return (common.b64_secret(key_pem), common.b64_secret(cert_pem))
        return (key_pem, cert_pem)

    def get_client_cert(
        self,
        client_component: Component,
        encode: bool = True,
    ) -> str:
        """Get the CA's public cert

        Args:
            client_component:  Component
                The Component that manages the client. This implementation does
                not need the component.
            encode:  bool
                Whether or not to base64 encode the output pem strings

        Returns:
            crt_pem:  Optional[str]
               The pem-encoded cert (base64 encoded if encode set)
        """
        log.debug2("Getting client cert for %s", client_component)
        _, ca_crt = self._component.get_ca_key_cert()
        if encode:
            return common.b64_secret(ca_crt)
        return ca_crt


## Component ###################################################################


class InternalCaComponent(Oper8xComponent):
    """This Component will manage a single secret containing a CA key/cert pair"""

    CA_SECRET_NAME = "infra-tls-ca"
    CA_KEY_FILENAME = "key.ca.pem"
    CA_CRT_FILENAME = "crt.ca.pem"

    name = "internal-tls"

    ## Component Interface #####################################################

    def __init__(
        self,
        session: Session,
        *args,
        labels: Optional[dict] = None,
        **kwargs,
    ):
        """Construct the parent Component and set up internal data holders"""
        super().__init__(*args, session=session, **kwargs)
        self._ca_key_pem = None
        self._ca_crt_pem = None

        # Pull labels from config or use defaults
        self._labels = labels

    def build_chart(self, *args, **kwargs):
        """Implement delayed chart construction in build_chart"""

        # Make sure the data values are populated
        self._initialize_data()

        # Get the labels to use for the secret
        secret_cluster_name = self._get_secret_name()
        labels = self._labels
        if labels is None:
            labels = common.get_labels(
                cluster_name=secret_cluster_name,
                session=self.session,
                component_name=self.CA_SECRET_NAME,
            )
        log.debug("Creating internal CA secret: %s", secret_cluster_name)
        self.add_resource(
            name=self.CA_SECRET_NAME,
            obj=dict(
                kind="Secret",
                apiVersion="v1",
                metadata=common.metadata_defaults(
                    session=self.session,
                    cluster_name=secret_cluster_name,
                    labels=labels,
                ),
                data={
                    self.CA_KEY_FILENAME: common.b64_secret(self._ca_key_pem),
                    self.CA_CRT_FILENAME: common.b64_secret(self._ca_crt_pem),
                },
            ),
        )

    ## Public Utilities ########################################################

    def get_ca_key_cert(self) -> Tuple[str, str]:
        """Get the pem-encoded CA key cert pair

        Returns:
            ca_key_pem:  str
                The pem-encoded (not base64 encoded) secret key
            ca_crt_pem:  str
                The pem-encoded (not base64 encoded) secret cert
        """
        self._initialize_data()
        return self._ca_key_pem, self._ca_crt_pem

    ## Implementation Details ##################################################

    def _get_secret_name(self) -> str:
        """Get the CA secret name with any scoping applied"""
        return self.get_cluster_name(self.CA_SECRET_NAME)

    def _initialize_data(self):
        """Initialize the data if needed"""

        # If this is the first time, actually do the init
        if None in [self._ca_crt_pem, self._ca_key_pem]:
            secret_cluster_name = self._get_secret_name()
            log.debug2("Cluster TLS Secret Name: %s", secret_cluster_name)
            success, content = self.session.get_object_current_state(
                kind="Secret",
                name=secret_cluster_name,
            )
            assert_cluster(
                success, f"Failed to check cluster for [{secret_cluster_name}]"
            )
            if content is not None:
                # Extract the pem strings
                key_pem = content.get("data", {}).get(self.CA_KEY_FILENAME)
                crt_pem = content.get("data", {}).get(self.CA_CRT_FILENAME)
                if None in [key_pem, crt_pem]:
                    log.warning(
                        "Found CA secret [%s] but content is invalid!",
                        secret_cluster_name,
                    )
                    self._generate()
                else:
                    log.debug("Found valid CA secret content.")
                    self._ca_key_pem = common.b64_secret_decode(key_pem)
                    self._ca_crt_pem = common.b64_secret_decode(crt_pem)
            else:
                log.debug2("No existing CA secret found. Generating.")
                self._generate()

    def _generate(self):
        """Generate a new CA"""
        key, self._ca_key_pem = tls.generate_key(encode=False)
        self._ca_crt_pem = tls.generate_ca_cert(key, encode=False)
